# -*- coding: utf-8 -*-

import pytest
from typing import Tuple

from gordo import parse_version, __version__


def test_version():
    assert isinstance(__version__, str)
    major, minor, is_unstable = parse_version(__version__)
    assert isinstance(major, int)
    assert isinstance(minor, int)
    assert isinstance(is_unstable, bool)


@pytest.mark.parametrize(
    "version,expected",
    [
        ("1.1.1", (1, 1, False)),
        ("1.1.1.dev-a1", (1, 1, True)),
        ("0.55.0-rc1", (0, 55, True)),
        ("0.0.0", (0, 0, True)),
    ],
)
def test_version_parser(version: str, expected: Tuple[int, int, int]):
    assert parse_version(version) == expected


def test_version_with_error():
    with pytest.raises(ValueError):
        parse_version("not_a_version")



import subprocess
import re

import dictdiffer

from typing import Optional
from packaging import version

_version_re = re.compile(r"^argo:\s+v?(.+)$")


class ArgoVersionError(Exception):
    pass


def patch_dict(original_dict: dict, patch_dictionary: dict) -> dict:
    """Patches a dict with another. Patching means that any path defines in the
    patch is either added (if it does not exist), or replaces the existing value (if
    it exists). Nothing is removed from the original dict, only added/replaced.

    Parameters
    ----------
    original_dict : dict
        Base dictionary which will get paths added/changed
    patch_dictionary: dict
        Dictionary which will be overlaid on top of original_dict

    Examples
    --------
    >>> patch_dict({"highKey":{"lowkey1":1, "lowkey2":2}}, {"highKey":{"lowkey1":10}})
    {'highKey': {'lowkey1': 10, 'lowkey2': 2}}
    >>> patch_dict({"highKey":{"lowkey1":1, "lowkey2":2}}, {"highKey":{"lowkey3":3}})
    {'highKey': {'lowkey1': 1, 'lowkey2': 2, 'lowkey3': 3}}
    >>> patch_dict({"highKey":{"lowkey1":1, "lowkey2":2}}, {"highKey2":4})
    {'highKey': {'lowkey1': 1, 'lowkey2': 2}, 'highKey2': 4}

    Returns
    -------
    dict
        A new dictionary which is the result of overlaying `patch_dictionary` on top of
        `original_dict`

    """
    diff = dictdiffer.diff(original_dict, patch_dictionary)
    adds_and_mods = [(f, d, s) for (f, d, s) in diff if f != "remove"]
    return dictdiffer.patch(adds_and_mods, original_dict)


def parse_argo_version(argo_version: str) -> Optional[version.Version]:
    """
    Try to parse Argo version.

    Parameters
    ----------
    argo_version: str

    Returns
    -------
        None if failed to parse.
    """
    parsed_version = version.parse(argo_version)
    if isinstance(parsed_version, version.Version):
        return parsed_version
    return None


def determine_argo_version(argo_binary: str = "argo") -> str:
    """
    Check installed Argo CLI version.

    Raises
    ------
    ArgoVersionError
        If something went wrong.
    Returns
    -------
        Version of installed argo version.
    """
    command = [argo_binary, "version", "--short"]
    message_suffix = ". Command: '%s'" % " ".join(command)
    try:
        result = subprocess.run(command, timeout=30, capture_output=True, check=True)
    except (subprocess.SubprocessError, FileNotFoundError):
        raise ArgoVersionError("Getting argo version exception" + message_suffix)
    if result is not None:
        stdout: Optional[str] = None
        try:
            stdout = result.stdout.decode("utf-8")
        except UnicodeDecodeError:
            raise ArgoVersionError(
                ("Unable to encode output %s" % str(stdout)) + message_suffix
            )
        if stdout is not None:
            m = _version_re.match(stdout.rstrip())
            if not m:
                raise ArgoVersionError(
                    ("Unable to parse version from %s" % str(stdout)) + message_suffix
                )
            return m[1]



#!/usr/bin/env python

import sys
import os
import re
import json
import argparse

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Optional, Tuple, Dict

if sys.version_info.major != 3 or sys.version_info.minor < 7:
    raise RuntimeError("Unsupported python version: %s" % sys.version)


re_tags = re.compile(r"^refs\/tags\/v?([^\/]*).*?$")
re_pull = re.compile(r"^refs\/pull\/(\d+).*?$")


class ImageType(Enum):
    dev = "dev"
    pr = "pr"
    prod = "prod"
    sha = "sha"


class Release(Enum):
    release = "release"
    prerelease = "prerelease"


def get_github_event(environ: Dict[str, str]):
    with open(environ["GITHUB_EVENT_PATH"], "r") as f:  # type: ignore
        return json.load(f)


@dataclass
class Settings:
    image_names: List[str]
    docker_image: str
    docker_prod_image: str
    base_image: str

    @classmethod
    def from_environ(cls, repository: str, image_names: List[str], environ=None):
        if environ is None:
            environ = os.environ
        if "DOCKER_REGISTRY" not in environ:
            raise RuntimeError("DOCKER_REGISTRY environment variable is empty")
        docker_image = environ["DOCKER_REGISTRY"] + "/" + repository
        docker_prod_image = ""
        if "DOCKER_PROD_REGISTRY" in environ:
            docker_prod_image = environ["DOCKER_PROD_REGISTRY"] + "/" + repository
        base_image = docker_image + "/base"
        return cls(
            image_names=image_names,
            docker_image=docker_image,
            docker_prod_image=docker_prod_image,
            base_image=base_image,
        )

    def get_docker_images(self, labels: List[str], for_prod: bool = False):
        docker_image = self.docker_prod_image if for_prod else self.docker_image
        if not docker_image:
            return
        for image_name in self.image_names:
            for label in labels:
                yield docker_image + "/" + image_name + ":" + label


@dataclass
class Context:
    image_type: ImageType
    version: str
    release: Optional[Release] = None


def prepare_context(environ=None):
    if environ is None:
        environ = os.environ
    image_type, version, release = ImageType.dev, "", None
    github_ref = environ["GITHUB_REF"]
    m = re_tags.match(github_ref)
    if m:
        version = m[1]
    if not version:
        m = re_pull.match(github_ref)
        if m:
            image_type, version = ImageType.pr, "pr-" + m[1]
    if environ["GITHUB_EVENT_NAME"] == "release":
        release = Release.prerelease
        prerelease = get_github_event(environ)["release"]["prerelease"]
        if not prerelease:
            release = Release.release
            image_type = ImageType.prod
    if not version:
        image_type, version = ImageType.sha, environ["GITHUB_SHA"][:8]
    return Context(image_type, version, release)


def version_labels(version: str) -> List[str]:
    labels = [version]
    splited = version.split(".")
    if len(splited) == 3:
        labels.append(splited[0])
        labels.append(splited[0] + "." + splited[1])
    return labels


def get_output_tags(settings: Settings, context: Context) -> List[str]:
    image_type = context.image_type
    tags = []
    if image_type is ImageType.pr:
        tags.extend(settings.get_docker_images([context.version]))
    else:
        tags.extend(settings.get_docker_images([context.version]))
        if context.release:
            tags.extend(settings.get_docker_images(["latest"]))
            if context.release is Release.release:
                dev_labels = version_labels(context.version)
                tags.extend(settings.get_docker_images(dev_labels))
        if image_type is ImageType.prod:
            tags.extend(settings.get_docker_images(["latest"], for_prod=True))
            prod_labels = version_labels(context.version)
            tags.extend(settings.get_docker_images(prod_labels, for_prod=True))
        if context.release is Release.release:
            tags.extend(settings.get_docker_images(["stable"]))
            if image_type is ImageType.prod:
                tags.extend(settings.get_docker_images(["stable"], for_prod=True))
    return tags


def bool_variable(value: bool):
    return "true" if value else ""


def get_variables(
    repository: str,
    image_names: List[str],
    with_pr: bool = False,
    with_sha: bool = False,
):
    variables: List[Tuple[str, str]] = []
    settings = Settings.from_environ(repository, image_names)
    push_image, login_cr, login_prod_cr = True, True, bool(settings.docker_prod_image)

    context = prepare_context()
    image_type = context.image_type
    if (image_type is ImageType.pr and not with_pr) or (
        image_type is ImageType.sha and not with_sha
    ):
        push_image, login_cr, login_prod_cr = False, False, False
    if image_type is not ImageType.prod:
        login_prod_cr = False
    tags = []
    if push_image:
        tags = get_output_tags(settings, context)

    variables = []
    variables.append(("version", context.version))
    variables.append(
        ("release_type", context.release.value if context.release is not None else "")
    )
    variables.append(("image_type", context.image_type.value))
    dt = datetime.utcnow()
    variables.append(("created", dt.strftime("%Y-%m-%dT%H:%M:%SZ")))
    variables.append(("base_image", settings.base_image + ":" + context.version))
    variables.append(("tags_gordo_base", ",".join(tags)))
    variables.append(("push_image", bool_variable(push_image)))
    variables.append(("login_cr", bool_variable(login_cr)))
    variables.append(("login_prod_cr", bool_variable(login_prod_cr)))
    return variables


def render_set_output(variables: List[Tuple[str, str]], environ=None):
    for name, value in variables:
        print("::set-output name=%s::%s" % (name, value))


def render_github_output(variables: List[Tuple[str, str]], environ=None):
    if environ is None:
        environ = os.environ
    with open(environ["GITHUB_OUTPUT"], "a") as f:
        for name, value in variables:
            f.write("%s=%s\n" % (name, value))


_renders = {
    "set-output": render_set_output,
    "github-output": render_github_output,
}
_default_render = "github-output"


def main():
    parser = argparse.ArgumentParser(
        description="Provides the variables such as docker images tags for GitHub workflow"
    )

    parser.add_argument(
        "-r",
        "--repository",
        required=True,
        help="Docker repository. Full url: <registry>/<repository>/<image_name>",
    )
    parser.add_argument(
        "-i", "--image-name", required=True, action="append", help="Docker image name"
    )
    parser.add_argument(
        "-p", "--with-pr", action="store_true", help="Run for PRs events"
    )
    parser.add_argument(
        "-s", "--with-sha", action="store_true", help="Run for non-releases"
    )
    parser.add_argument(
        "-e",
        "--render",
        choices=list(_renders.keys()),
        default=_default_render,
        help="Variables render type. Default: '%s'" % _default_render,
    )

    args = parser.parse_args()
    render = _renders[args.render]

    variables = get_variables(
        repository=args.repository,
        image_names=args.image_name,
        with_pr=args.with_pr,
        with_sha=args.with_sha,
    )
    render(variables)


if __name__ == "__main__":
    main()



import logging
import traceback
from packaging import version
from typing import Tuple
import warnings

DEFAULT_VERSION = "0.0.0"

try:
    from ._version import version as __version__
except ImportError:
    __version__ = DEFAULT_VERSION


def parse_version(input_version: str) -> Tuple[int, int, bool]:
    """
    Takes a string which starts with standard major.minor.patch.
    and returns the split of major and minor version as integers

    Parameters
    ----------
    input_version: str
        The semantic version string

    Returns
    -------
    Tuple[int, int, bool]
        major and minor versions, and flag "is this unstable version?"
    """
    parsed_version = version.parse(input_version)
    result: Tuple[int, int, bool] = (0, 0, True)
    if isinstance(parsed_version, version.Version):
        is_unstable = parsed_version.is_devrelease or parsed_version.is_prerelease
        result = parsed_version.major, parsed_version.minor, is_unstable
    else:
        try:
            split_version = input_version.split(".")[:2]
            major, minor = int(split_version[0]), int(split_version[1])
        except ValueError:
            raise ValueError("Malformed version %s" % input_version)
        result = major, minor, True
    if input_version == DEFAULT_VERSION:
        return result[0], result[1], True
    return result


MAJOR_VERSION, MINOR_VERSION, IS_UNSTABLE_VERSION = parse_version(__version__)

try:
    # FIXME(https://github.com/abseil/abseil-py/issues/99)
    # FIXME(https://github.com/abseil/abseil-py/issues/102)
    # Unfortunately, many libraries that include absl (including Tensorflow)
    # will get bitten by double-logging due to absl's incorrect use of
    # the python logging library:
    #   2019-07-19 23:47:38,829 my_logger   779 : test
    #   I0719 23:47:38.829330 139904865122112 foo.py:63] test
    #   2019-07-19 23:47:38,829 my_logger   779 : test
    #   I0719 23:47:38.829469 139904865122112 foo.py:63] test
    # The code below fixes this double-logging.  FMI see:
    #   https://github.com/tensorflow/tensorflow/issues/26691#issuecomment-500369493

    import absl.logging

    logging.root.removeHandler(absl.logging._absl_handler)
    absl.logging._warn_preinit_stderr = False

except Exception:
    warnings.warn(f"Failed to fix absl logging bug {traceback.format_exc()}")



# -*- coding: utf-8 -*-

from typing import Union

import pandas as pd
import numpy as np
from numpy.ma import masked_invalid

from sklearn.base import TransformerMixin


class InfImputer(TransformerMixin):
    def __init__(
        self,
        inf_fill_value=None,
        neg_inf_fill_value=None,
        strategy="minmax",
        delta: float = 2.0,
    ):
        """
        Fill inf/-inf values of a 2d array/dataframe with imputed or provided values
        By default it will find the min and max of each feature/column and fill -infs/infs
        with those values +/- ``delta``

        Parameters
        ----------
        inf_fill_value: numeric
            Value to fill 'inf' values
        neg_inf_fill_value: numeric
            Value to fill '-inf' values
        strategy: str
            How to fill values, irrelevant if fill value is provided.
            choices: 'extremes', 'minmax'
            -'extremes' will use the min and max values for the current datatype.
            such that 'inf' in a float32 dataset will have float32's largest value inserted.
            - 'minmax' will look at the min and max values in the feature where the -inf / inf
            appears and fill with the max/min found in that feature.
        delta: float
            Only applicable if ``strategy='minmax'``
            Will add/subtract the max/min value, by feature, by this delta. If the max value
            in a feature was 10 and ``delta=2`` any inf value will be filled with 12.
            Likewise, if the min feature was -10 any -inf will be filled with -12.
        """
        self.inf_fill_value = inf_fill_value
        self.neg_inf_fill_value = neg_inf_fill_value
        self.strategy = strategy
        self.delta = delta

    def get_params(self, deep=True):
        return {
            "inf_fill_value": self.inf_fill_value,
            "neg_inf_fill_value": self.neg_inf_fill_value,
            "strategy": self.strategy,
            "delta": self.delta,
        }

    def fit(self, X: Union[pd.DataFrame, np.ndarray], y=None):

        # Store the min/max for features in training.
        if self.strategy == "minmax":

            data = pd.DataFrame(X)  # ensure a dataframe

            # Calculate max/min allowable values
            max_allowable_value = np.finfo(data.values.dtype).max
            min_allowable_value = np.finfo(data.values.dtype).min

            # Get the max/min values in each feature, ignoring infs
            _posinf_fill_values = data.apply(lambda col: masked_invalid(col).max())
            _neginf_fill_values = data.apply(lambda col: masked_invalid(col).min())

            # Calculate a 1d arrays of fill values for each feature
            self._posinf_fill_values = _posinf_fill_values.apply(
                lambda val: val + self.delta
                if max_allowable_value - self.delta > val
                else max_allowable_value
            )
            self._neginf_fill_values = _neginf_fill_values.apply(
                lambda val: val - self.delta
                if min_allowable_value + self.delta < val
                else min_allowable_value
            )

        return self

    def transform(self, X: Union[pd.DataFrame, np.ndarray], y=None):

        # Ensure we're dealing with numpy array if it's a dataframe or similar
        X = X.values if isinstance(X, pd.DataFrame) else X

        # Apply specific fill values if provided.
        if self.inf_fill_value is not None:
            X[np.isposinf(X)] = self.inf_fill_value
        if self.neg_inf_fill_value is not None:
            X[np.isneginf(X)] = self.neg_inf_fill_value

        # May still be left over infs, if only one fill value was supplied for example
        if self.strategy is not None:
            return getattr(self, f"_fill_{self.strategy}")(X)
        return X

    def _fill_extremes(self, X: np.ndarray):
        """
        Fill negative and postive infs with their dtype's min/max values
        """
        X[np.isposinf(X)] = np.finfo(X.dtype).max
        X[np.isneginf(X)] = np.finfo(X.dtype).min
        return X

    def _fill_minmax(self, X: np.ndarray):
        """
        Fill inf/-inf values in features of the array based on their min & max values.
        Compounded by the ``power`` value so long as the result doesn't exceed the
        current array's dtype's max/min. Otherwise it will use those.
        """

        # For each feature fill inf/-inf with pre-calculate fill values
        for feature_idx, (posinf_fill, neginf_fill) in enumerate(
            zip(self._posinf_fill_values, self._neginf_fill_values)
        ):
            X[:, feature_idx][np.isposinf(X[:, feature_idx])] = posinf_fill
            X[:, feature_idx][np.isneginf(X[:, feature_idx])] = neginf_fill
        return X



# -*- coding: utf-8 -*-

import logging
import functools
import zlib
import os
import io
import pickle
import re
import shutil

import dateutil
import timeit
from datetime import datetime
from typing import Union, List

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from flask import request, g, jsonify, make_response, Response
from functools import lru_cache, wraps
from sklearn.base import BaseEstimator
from werkzeug.exceptions import NotFound, UnprocessableEntity, InternalServerError

from gordo import serializer

from .properties import get_tags, get_target_tags

"""
Tools used between different views

Namely :func:`.extract_X_y` and :func:`.base_dataframe` decorators which, when
wrapping methods will perform the tasks to extracting X & y, and creating the 
basic dataframe output, respectively.
"""

logger = logging.getLogger(__name__)

gordo_name_re = re.compile(r"^[a-zA-Z\d-]+")
revision_re = re.compile(r"^\d+$")


def validate_revision(revision: str) -> bool:
    return bool(revision_re.match(revision))


def dataframe_into_parquet_bytes(
    df: pd.DataFrame, compression: str = "snappy"
) -> bytes:
    """
    Convert a dataframe into bytes representing a parquet table.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to be compressed
    compression: str
        Compression to use, passed to  :func:`pyarrow.parquet.write_table`

    Returns
    -------
    bytes
    """
    table = pa.Table.from_pandas(df)
    buf = pa.BufferOutputStream()
    pq.write_table(table, buf, compression=compression)
    return buf.getvalue().to_pybytes()


def dataframe_from_parquet_bytes(buf: bytes) -> pd.DataFrame:
    """
    Convert bytes representing a parquet table into a pandas dataframe.

    Parameters
    ----------
    buf: bytes
        Bytes representing a parquet table. Can be the direct result from
        `func`::gordo.server.utils.dataframe_into_parquet_bytes

    Returns
    -------
    pandas.DataFrame
    """
    table = pq.read_table(io.BytesIO(buf))
    return table.to_pandas()


def dataframe_to_dict(df: pd.DataFrame) -> dict:
    """
    Convert a dataframe can have a :class:`pandas.MultiIndex` as columns into a dict
    where each key is the top level column name, and the value is the array
    of columns under the top level name. If it's a simple dataframe, :meth:`pandas.core.DataFrame.to_dict`
    will be used.

    This allows :func:`json.dumps` to be performed, where :meth:`pandas.DataFrame.to_dict()`
    would convert such a multi-level column dataframe into keys of ``tuple`` objects, which are
    not json serializable. However this ends up working with :meth:`pandas.DataFrame.from_dict`

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe expected to have columns of type :class:`pandas.MultiIndex` 2 levels deep.

    Returns
    -------
    List[dict]
        List of records representing the dataframe in a 'flattened' form.


    Examples
    --------
    >>> import pprint
    >>> import pandas as pd
    >>> import numpy as np
    >>> columns = pd.MultiIndex.from_tuples((f"feature{i}", f"sub-feature-{ii}") for i in range(2) for ii in range(2))
    >>> index = pd.date_range('2019-01-01', '2019-02-01', periods=2)
    >>> df = pd.DataFrame(np.arange(8).reshape((2, 4)), columns=columns, index=index)
    >>> df  # doctest: +NORMALIZE_WHITESPACE
                    feature0                    feature1
               sub-feature-0 sub-feature-1 sub-feature-0 sub-feature-1
    2019-01-01             0             1             2             3
    2019-02-01             4             5             6             7
    >>> serialized = dataframe_to_dict(df)
    >>> pprint.pprint(serialized)
    {'feature0': {'sub-feature-0': {'2019-01-01': 0, '2019-02-01': 4},
                  'sub-feature-1': {'2019-01-01': 1, '2019-02-01': 5}},
     'feature1': {'sub-feature-0': {'2019-01-01': 2, '2019-02-01': 6},
                  'sub-feature-1': {'2019-01-01': 3, '2019-02-01': 7}}}

    """
    # Need to copy, because Python's mutability allowed .index assignment to mutate the passed df
    data = df.copy()
    if isinstance(data.index, pd.DatetimeIndex):
        data.index = data.index.astype(str)
    if isinstance(df.columns, pd.MultiIndex):
        return {
            col: data[col].to_dict()
            if isinstance(data[col], pd.DataFrame)
            else pd.DataFrame(data[col]).to_dict()
            for col in data.columns.get_level_values(0)
        }
    else:
        return data.to_dict()


def dataframe_from_dict(data: dict) -> pd.DataFrame:
    """
    The inverse procedure done by :func:`.multi_lvl_column_dataframe_from_dict`
    Reconstructed a MultiIndex column dataframe from a previously serialized one.

    Expects ``data`` to be a nested dictionary where each top level key has a value
    capable of being loaded from :func:`pandas.core.DataFrame.from_dict`

    Parameters
    ----------
    data: dict
        Data to be loaded into a MultiIndex column dataframe

    Returns
    -------
    pandas.core.DataFrame
        MultiIndex column dataframe.

    Examples
    --------
    >>> serialized = {
    ... 'feature0': {'sub-feature-0': {'2019-01-01': 0, '2019-02-01': 4},
    ...              'sub-feature-1': {'2019-01-01': 1, '2019-02-01': 5}},
    ... 'feature1': {'sub-feature-0': {'2019-01-01': 2, '2019-02-01': 6},
    ...              'sub-feature-1': {'2019-01-01': 3, '2019-02-01': 7}}
    ... }
    >>> dataframe_from_dict(serialized)  # doctest: +NORMALIZE_WHITESPACE
                    feature0                    feature1
           sub-feature-0 sub-feature-1 sub-feature-0 sub-feature-1
    2019-01-01             0             1             2             3
    2019-02-01             4             5             6             7
    """

    if isinstance(data, dict) and any(isinstance(val, dict) for val in data.values()):
        try:
            keys = data.keys()
            df: pd.DataFrame = pd.concat(
                (pd.DataFrame.from_dict(data[key]) for key in keys), axis=1, keys=keys
            )
        except (ValueError, AttributeError):
            df = pd.DataFrame.from_dict(data)
    else:
        df = pd.DataFrame.from_dict(data)

    try:
        df.index = df.index.map(dateutil.parser.isoparse)  # type: ignore
    except (TypeError, ValueError):
        df.index = df.index.map(int)

    df.sort_index(inplace=True)

    return df


def parse_iso_datetime(datetime_str: str) -> datetime:
    parsed_date = dateutil.parser.isoparse(datetime_str)  # type: ignore
    if parsed_date.tzinfo is None:
        raise ValueError(
            f"Provide timezone to timestamp {datetime_str}."
            f" Example: for UTC timezone use {datetime_str + 'Z'} or {datetime_str + '+00:00'} "
        )
    return parsed_date


def _verify_dataframe(
    df: pd.DataFrame, expected_columns: List[str]
) -> Union[Response, pd.DataFrame]:
    """
    Verify the dataframe, setting the column names to ``expected_columns``
    if not already labeled and the length of the columns match the length of the expected columns.

    If it fails, it will return an instance of :class:`flask.wrappers.Response`

    Parameters
    ----------
    df: pandas.core.DataFrame
        DataFrame to verify.
    expected_columns: List[str]
        List of expected column names to give if the dataframe does not consist of them
        but the number of columns matches ``len(expected_columns)``

    Returns
    -------
    Union[flask.wrappers.Response, pandas.core.DataFrame]
    """
    if not isinstance(df.columns, pd.MultiIndex):
        if not all(col in df.columns for col in expected_columns):

            # If the length doesn't mach, then we can't reliably determine what data we have hre.
            if len(df.columns) != len(expected_columns):
                msg = dict(
                    message=f"Unexpected features: "
                    f"was expecting {expected_columns} length of {len(expected_columns)}, "
                    f"but got {df.columns} length of {len(df.columns)}"
                )
                return make_response((jsonify(msg), 400))

            # Otherwise we were send a list/ndarray data format which we assume the client has
            # ordered correctly to the order of the expected_columns.
            else:
                df.columns = expected_columns

        # All columns exist in the dataframe, select them which thus ensures order and removes extra columns
        else:
            df = df[expected_columns]
        return df
    else:
        msg = {
            "message": f"Server does not support multi-level dataframes at this time: {df.columns.tolist()}"
        }
        return make_response((jsonify(msg), 400))


def extract_X_y(method):
    """
    For a given flask view, will attempt to extract an 'X' and 'y' from
    the request and assign it to flask's 'g' global request context

    If it fails to extract 'X' and (optionally) 'y' from the request, it will **not** run the
    function but return a ``BadRequest`` response notifying the client of the failure.

    Parameters
    ----------
    method: Callable
        The flask route to decorate, and will return it's own response object
        and will want to use ``flask.g.X`` and/or ``flask.g.y``

    Returns
    -------
    flask.Response
        Will either run a :class:`flask.Response` with status code 400 if it fails
        to extract the X and optionally the y. Otherwise will run the decorated ``method``
        which is also expected to return some sort of :class:`flask.Response` object.
    """

    @functools.wraps(method)
    def wrapper_method(*args, **kwargs):
        start_time = timeit.default_timer()
        # Data provided by the client
        if request.method == "POST":

            # Always require an X, be it in JSON or file/parquet format.
            if request.is_json:
                if "X" not in (request.json or {}):
                    message = dict(message='Cannot predict without "X"')
                    return make_response((jsonify(message), 400))
            else:
                if "X" not in request.files:
                    message = dict(message='Cannot predict without "X"')
                    return make_response((jsonify(message), 400))

            if request.is_json:
                X = dataframe_from_dict(request.json["X"])
                y = request.json.get("y")
                if y is not None:
                    y = dataframe_from_dict(y)
            else:
                X = dataframe_from_parquet_bytes(request.files["X"].read())
                y = request.files.get("y")
                if y is not None:
                    y = dataframe_from_parquet_bytes(y.read())

            X = _verify_dataframe(X, [t.name for t in get_tags()])

            # Verify y if it's not None
            if y is not None:
                y = _verify_dataframe(y, [t.name for t in get_target_tags()])

            # If either X or y came back as a Response type, there was an error
            for data_or_resp in [X, y]:
                if isinstance(data_or_resp, Response):
                    return data_or_resp
        else:
            raise NotImplementedError(
                f"Cannot extract X and y from '{request.method}' request."
            )

        # Assign X and y to the request's global context
        g.X, g.y = X, y

        try:
            logger.debug(f"Size of X: {X.size}, size of y: {y.size}")
        except AttributeError:
            logger.debug(f"Size of X: {X.size}, y is None")
        logger.debug(f"Time to parse X and y: {timeit.default_timer() - start_time}s")

        # And run the original method.
        return method(*args, **kwargs)

    return wrapper_method


@lru_cache(maxsize=int(os.getenv("N_CACHED_MODELS", 2)))
def load_model(directory: str, name: str) -> BaseEstimator:
    """
    Load a given model from the directory by name.

    Parameters
    ----------
    directory: str
        Directory to look for the model
    name: str
        Name of the model to load, this would be the sub directory within the
        directory parameter.

    Returns
    -------
    BaseEstimator
    """
    start_time = timeit.default_timer()
    model = serializer.load(os.path.join(directory, name))
    logger.debug(f"Time to load model: {timeit.default_timer() - start_time}s")
    return model


def check_metadata_file(directory: str, name: str):
    """
    Checking if the directory with metadata exists since it might be deleted through DELETE endpoint
    """
    # TODO consider using https://pypi.org/project/ring/ with the ability to delete items from lru cache
    full_model_dir = os.path.join(directory, name)
    if not serializer.metadata_path(full_model_dir):
        raise FileNotFoundError("Unable to load metadata.json file")


def load_metadata(directory: str, name: str) -> dict:
    """
    Load metadata from a directory for a given model by name.

    Parameters
    ----------
    directory: str
        Directory to look for the model's metadata
    name: str
        Name of the model to load metadata for, this would be the sub directory
        within the directory parameter.

    Returns
    -------
    dict
    """
    compressed_metadata = _load_compressed_metadata(directory, name)
    return pickle.loads(zlib.decompress(compressed_metadata))


_n_cached_metadata = int(os.getenv("N_CACHED_METADATA", 250))


@lru_cache(maxsize=_n_cached_metadata)
def _load_compressed_metadata(directory: str, name: str):
    """
    Loads the metadata for model 'name' from directory 'directory', and returns it as a
    zlib compressed pickle, to use as little space as possible in the cache.

    Notes
    ----
    Some simple measurement indicated that a typical metadata dict uses 37kb in memory,
    while pickled it uses 8kb, and pickled-compressed it uses 4kb.

    """
    metadata = serializer.load_metadata(os.path.join(directory, name))
    return zlib.compress(pickle.dumps(metadata))


def delete_revision(directory: str, name: str):
    """
    Delete model revision

    Parameters
    ----------
    directory - Revision directory
    name - Model name
    """
    full_path = os.path.join(directory, name)
    if not os.path.isfile(os.path.join(full_path, "metadata.json")):
        raise NotFound("Not found")
    shutil.rmtree(full_path, ignore_errors=True)
    if os.path.exists(full_path):
        raise InternalServerError("Unable to delete this model revision folder")
    if not os.listdir(directory):
        shutil.rmtree(directory, ignore_errors=True)
        if os.path.exists(directory):
            raise InternalServerError("Unable to delete this revision folder")


def validate_gordo_name(gordo_name: str):
    """
    gordo_name argument should contains alpha-numericals or '-' symbols
    """
    if gordo_name and not gordo_name_re.match(gordo_name):
        raise UnprocessableEntity("gordo_name field has wrong format")


@lru_cache(maxsize=_n_cached_metadata)
def load_info(directory: str, name: str) -> dict:
    # TODO better docstring
    return serializer.load_info(os.path.join(directory, name))


def metadata_required(f):
    """
    Decorate a view which has ``gordo_name`` as a url parameter and will
    set ``g.metadata`` to that model's metadata
    """

    @wraps(f)
    def wrapper(*args: tuple, gordo_project: str, gordo_name: str, **kwargs: dict):
        validate_gordo_name(gordo_name)
        g.info = {}
        try:
            g.info = load_info(directory=g.collection_dir, name=gordo_name)
        except FileNotFoundError:
            pass
        try:
            check_metadata_file(g.collection_dir, gordo_name)
            g.metadata = load_metadata(directory=g.collection_dir, name=gordo_name)
        except FileNotFoundError:
            raise NotFound(f"No metadata found for '{gordo_name}'")
        else:
            return f(*args, **kwargs)

    return wrapper


def model_required(f):
    """
    Decorate a view which has ``gordo_name`` as a url parameter and will
    set ``g.model`` to be the loaded model and ``g.metadata``
    to that model's metadata
    """

    @wraps(f)
    def wrapper(*args: tuple, gordo_project: str, gordo_name: str, **kwargs: dict):
        validate_gordo_name(gordo_name)
        try:
            check_metadata_file(g.collection_dir, gordo_name)
            g.model = load_model(directory=g.collection_dir, name=gordo_name)
        except FileNotFoundError:
            raise NotFound(f"No such model found: '{gordo_name}'")
        else:
            # If the model was required, the metadata is also required.
            return metadata_required(f)(
                *args, gordo_project=gordo_project, gordo_name=gordo_name, **kwargs
            )

    return wrapper



# -*- coding: utf-8 -*-

import os
import io
import logging
import traceback
import timeit

import pandas as pd
from flask import Blueprint, current_app, g, send_file, make_response, jsonify, request

from gordo import __version__, serializer
from gordo.server import utils as server_utils
from gordo.machine.model import utils as model_utils
from gordo.server.utils import (
    validate_gordo_name,
    validate_revision,
    delete_revision,
)
from gordo.server.properties import get_tags, get_target_tags
from gordo.server import model_io
from typing import Any


logger = logging.getLogger(__name__)

base_blueprint = Blueprint("base_model_view", __name__)


@base_blueprint.route(
    "/gordo/v0/<gordo_project>/<gordo_name>/prediction", methods=["POST"]
)
@server_utils.model_required
@server_utils.extract_X_y
def post_prediction():
    """
    Process a POST request by using provided user data

    A typical response might look like this

    .. code-block:: python

        {
            'data': [
                {
                    'end': ['2016-01-01T00:10:00+00:00'],
                    'model-output': [0.0005317790200933814,
                                     -0.0001525811239844188,
                                     0.0008310950361192226,
                                     0.0015755111817270517],
                    'original-input': [0.9135588550070414,
                                       0.3472517774179448,
                                       0.8994921857179736,
                                       0.11982773108991263],
                    'start': ['2016-01-01T00:00:00+00:00'],
                },
                ...
            ],

            'tags': [
                {'asset': None, 'name': 'tag-0'},
                {'asset': None, 'name': 'tag-1'},
                {'asset': None, 'name': 'tag-2'},
                {'asset': None, 'name': 'tag-3'}
            ],
            'time-seconds': '0.1937'
        }
    """
    context: dict[Any, Any] = dict()
    X = g.X
    process_request_start_time_s = timeit.default_timer()

    try:
        output = model_io.get_model_output(model=g.model, X=X)
    except ValueError as err:
        tb = traceback.format_exc()
        logger.error(
            f"Failed to predict or transform; error: {err} - \nTraceback: {tb}"
        )
        context["error"] = f"ValueError: {str(err)}"
        return make_response((jsonify(context), 400))

    # Model may only be a transformer, probably an AttributeError, but catch all to avoid logging other
    # exceptions twice if it happens.
    except Exception as exc:
        tb = traceback.format_exc()
        logger.error(
            f"Failed to predict or transform; error: {exc} - \nTraceback: {tb}"
        )
        context["error"] = "Something unexpected happened; check your input data"
        return make_response((jsonify(context), 400))

    else:
        get_model_output_time_s = timeit.default_timer()
        logger.debug(
            f"Calculating model output took "
            f"{get_model_output_time_s-process_request_start_time_s} s"
        )
        data = model_utils.make_base_dataframe(
            tags=get_tags(),
            model_input=X.values if isinstance(X, pd.DataFrame) else X,
            model_output=output,
            target_tag_list=get_target_tags(),
            index=X.index,
        )
        if request.args.get("format") == "parquet":
            return send_file(
                io.BytesIO(server_utils.dataframe_into_parquet_bytes(data)),
                mimetype="application/octet-stream",
            )
        else:
            context["data"] = server_utils.dataframe_to_dict(data)
            return make_response((jsonify(context), context.pop("status-code", 200)))


@base_blueprint.route(
    "/gordo/v0/<gordo_project>/<gordo_name>/revision/<revision>", methods=["DELETE"]
)
def delete_model_revision(gordo_name: str, revision: str, **kwargs):
    """
    Delete provided model revision from the disk.
    """
    validate_gordo_name(gordo_name)
    if not validate_revision(revision):
        return make_response(
            (jsonify({"error": "Revision should only contains numbers."}), 422)
        )
    if revision == g.current_revision:
        return make_response(
            (jsonify({"error": "Unable to delete current revision."}), 409)
        )
    revision_dir = os.path.join(g.collection_dir, "..", revision)
    delete_revision(revision_dir, gordo_name)
    return make_response(jsonify({"ok": True}), 200)


@base_blueprint.route(
    "/gordo/v0/<gordo_project>/<gordo_name>/metadata", methods=["GET"]
)
@base_blueprint.route(
    "/gordo/v0/<gordo_project>/<gordo_name>/healthcheck", methods=["GET"]
)
@server_utils.metadata_required
def get_metadata():
    """
    Serve model / server metadata

    Get metadata about this endpoint, also serves as /healthcheck endpoint
    """
    model_collection_env_var = current_app.config["MODEL_COLLECTION_DIR_ENV_VAR"]
    metadata = {}
    if g.info:
        metadata = g.info
    metadata.update(
        {
            "gordo-server-version": __version__,
            "metadata": g.metadata,
            "env": {model_collection_env_var: os.environ.get(model_collection_env_var)},
        }
    )
    return metadata


@base_blueprint.route(
    "/gordo/v0/<gordo_project>/<gordo_name>/download-model", methods=["GET"]
)
@server_utils.model_required
def get_download_model():
    """
    Download the trained model

    Responds with a serialized copy of the current model being served.

    Returns
    -------
    bytes
        Results from ``gordo.serializer.dumps()``
    """
    serialized_model = serializer.dumps(g.model)
    buff = io.BytesIO(serialized_model)
    return send_file(buff, download_name="model.pickle")


@base_blueprint.route("/gordo/v0/<gordo_project>/models", methods=["GET"])
def get_model_list(gordo_project: str):
    """
    List the current models capable of being served by the server
    """
    available_models = []
    try:
        available_models = os.listdir(g.collection_dir)
    except FileNotFoundError:
        available_models = []
    finally:
        return jsonify({"models": available_models})


@base_blueprint.route("/gordo/v0/<gordo_project>/revisions", methods=["GET"])
def get_revision_list(gordo_project: str):
    """
    List the available revisions the model can serve.
    """
    try:
        available_revisions = os.listdir(os.path.join(g.collection_dir, ".."))
    except FileNotFoundError:
        logger.error(
            f"Attempted to list directories above {g.collection_dir} but failed with: {traceback.format_exc()}"
        )
        available_revisions = [g.current_revision]
    return jsonify(
        {"latest": g.current_revision, "available-revisions": available_revisions}
    )


@base_blueprint.route("/gordo/v0/<gordo_project>/expected-models", methods=["GET"])
def get(gordo_project: str):
    return jsonify({"expected-models": current_app.config["EXPECTED_MODELS"]})



import pytest
import peewee
from gordo.reporters.postgres import PostgresReporter, Machine as PostgresMachine
from gordo.machine import Machine
from gordo.reporters.exceptions import ReporterException


def test_postgres_reporter(postgresdb, metadata):
    """
    Check logging  of a machine into postgres
    """
    reporter1 = PostgresReporter(host="localhost")
    machine1 = Machine.from_dict(metadata)

    # Before inserting, the machine does not exist.
    with pytest.raises(peewee.DoesNotExist):
        PostgresMachine.get(PostgresMachine.name == machine1.name)

    reporter1.report(machine1)

    record = PostgresMachine.get(PostgresMachine.name == machine1.name)
    assert record.name == machine1.name

    # Create another logger to ensure nothing happened to the DB
    reporter2 = PostgresReporter(host="localhost")
    machine2 = Machine.from_dict(metadata)
    machine2.name = "another-machine"

    reporter2.report(machine2)

    # The first machine is still there
    record = PostgresMachine.get(PostgresMachine.name == machine1.name)
    assert record.name == machine1.name

    # And the second
    record = PostgresMachine.get(PostgresMachine.name == machine2.name)
    assert record.name == machine2.name


def test_postgres_exceptions(metadata, postgresdb):

    # No database to connect to
    with pytest.raises(ReporterException):
        PostgresReporter(host="does-not-exist")

    # Bad machine to report
    reporter = PostgresReporter(host="localhost")
    with pytest.raises(ReporterException):
        reporter.report("This it not a a machine instance.")  # type: ignore


def test_overwrite_report(postgresdb, metadata):
    """
    Ensure saving same machine is ok.
    """
    reporter1 = PostgresReporter(host="localhost")
    reporter2 = PostgresReporter(host="localhost")

    machine1 = Machine.from_dict(metadata)
    machine2 = Machine.from_dict(metadata)

    reporter1.report(machine1)

    # Reporting twice should be ok.
    reporter2.report(machine2)

    results = PostgresMachine.select().where(PostgresMachine.name == machine1.name)
    assert len([result for result in results]) == 1



# -*- coding: utf-8 -*-

import io
import logging
import timeit
import typing

from flask import Blueprint, make_response, jsonify, g, request, send_file

from gordo.server import utils, properties

logger = logging.getLogger(__name__)

anomaly_blueprint = Blueprint("ioc_anomaly_blueprint", __name__)


DELETED_FROM_RESPONSE_COLUMNS = (
    "smooth-tag-anomaly-scaled",
    "smooth-total-anomaly-scaled",
    "smooth-tag-anomaly-unscaled",
    "smooth-total-anomaly-unscaled",
)


def _create_anomaly_response(start_time: float = None):
    """
    Use the current ``X`` and ``y`` to create an anomaly specific response
    using the trained ML model's ``.anomaly()`` method.

    Parameters
    ----------
    start_time: Optional[float]
        Start time to use when timing the processing time of the request, will construct a new
        one if not provided.

    Returns
    -------
    flask.Response
        The formatted anomaly representation response object.
    """
    if start_time is None:
        start_time = timeit.default_timer()

    # To use this endpoint, we need a 'y' to calculate the errors.
    if g.y is None:
        message = {"message": "Cannot perform anomaly without 'y' to compare against."}
        return make_response((jsonify(message), 400))

    # Now create an anomaly dataframe from the base response dataframe
    try:
        anomaly_df = g.model.anomaly(g.X, g.y, frequency=properties.get_frequency())
    except AttributeError:
        msg = {
            "message": f"Model is not an AnomalyDetector, it is of type: {type(g.model)}"
        }
        return make_response(jsonify(msg), 422)  # 422 Unprocessable Entity

    if request.args.get("all_columns") is None:
        columns_for_delete = []
        for column in anomaly_df:
            if column[0] in DELETED_FROM_RESPONSE_COLUMNS:
                columns_for_delete.append(column)
        anomaly_df = anomaly_df.drop(columns=columns_for_delete)

    if request.args.get("format") == "parquet":
        return send_file(
            io.BytesIO(utils.dataframe_into_parquet_bytes(anomaly_df)),
            mimetype="application/octet-stream",
        )
    else:
        context: typing.Dict[typing.Any, typing.Any] = dict()
        context["data"] = utils.dataframe_to_dict(anomaly_df)
        context["time-seconds"] = f"{timeit.default_timer() - start_time:.4f}"
        return make_response(jsonify(context), context.pop("status-code", 200))


@anomaly_blueprint.route(
    "/gordo/v0/<gordo_project>/<gordo_name>/anomaly/prediction", methods=["POST"]
)
@utils.model_required
@utils.extract_X_y
def post_anomaly_prediction():
    """
    Serve model predictions via POST method.

    Gives back predictions looking something like this
    (depending on anomaly model being served)::

        {
        'data': [
            {
           'end': ['2016-01-01T00:10:00+00:00'],
           'tag-anomaly-scaled': [0.913027075986948,
                                  0.3474043585419292,
                                  0.8986610906818544,
                                  0.11825221990818557],
           'tag-anomaly-unscaled': [10.2335327305725986948,
                                  4.234343958392+3293,
                                  10.379394390232232,
                                  3.32093438982743929],
           'model-output': [0.0005317790200933814,
                            -0.0001525811239844188,
                            0.0008310950361192226,
                            0.0015755111817270517],
           'original-input': [0.9135588550070414,
                              0.3472517774179448,
                              0.8994921857179736,
                              0.11982773108991263],
           'start': ['2016-01-01T00:00:00+00:00'],
           'total-anomaly-unscaled': [1.3326228173185086],
           'total-anomaly-scaled': [0.3020328328002392],
            },
            ...
        ],

     'tags': [{'asset': None, 'name': 'tag-0'},
              {'asset': None, 'name': 'tag-1'},
              {'asset': None, 'name': 'tag-2'},
              {'asset': None, 'name': 'tag-3'}],
     'time-seconds': '0.1937'}
    """
    start_time = timeit.default_timer()
    return _create_anomaly_response(start_time)



from collections import OrderedDict
from typing import Optional, List, Dict, cast, Union

from gordo_core.metadata import sensor_tags_from_build_metadata
from gordo_core.sensor_tag import (
    SensorTag,
    Tag,
    normalize_sensor_tag,
    extract_tag_name,
)

TagsList = List[Union[Dict, List, str, SensorTag]]


def normalize_sensor_tags(
    build_dataset_metadata: dict, tag_list: TagsList, **kwargs: Optional[str]
) -> List[SensorTag]:
    """
    Load tag information from the metadata

    Parameters
    ----------
    build_dataset_metadata: dict
        build_metadata.dataset part of the metadata
    tag_list: TagsList
        Tag list that needs to be loaded
    kwargs: Optional[str]
        Additional fields for `normalize_sensor_tag()`

    Returns
    -------
    List[SensorTag]

    """
    tags: Dict[str, Tag] = OrderedDict()
    for tag in tag_list:
        tag_name: str
        if type(tag) is str:
            tags[cast(str, tag)] = tag
        else:
            normalized_tag = normalize_sensor_tag(tag, **kwargs)
            tag_name = extract_tag_name(normalized_tag)
            tags[tag_name] = normalized_tag
    normalized_sensor_tags = sensor_tags_from_build_metadata(
        build_dataset_metadata, set(tags.keys())
    )
    normalized_tag_list: List[SensorTag] = []
    for tag_name in tags.keys():
        normalized_tag_list.append(normalized_sensor_tags[tag_name])
    return normalized_tag_list


def join_json_paths(element: str, json_path: str = None) -> str:
    """
    Join two JSON paths

    Examples
    --------
    >>> join_json_paths("machines[0]", "spec.config")
    'spec.config.machines[0]'
    >>> join_json_paths("dataset")
    'dataset'

    Parameters
    ----------
    element: str
        Element name
    json_path: str
        Root JSON path

    Returns
    -------

    """
    json_paths = []
    if json_path:
        json_paths.append(json_path)
    json_paths.append(element)
    return ".".join(json_paths)



import json
import yaml

from typing import Any
from datetime import datetime
from gordo_core.sensor_tag import SensorTag

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f+%z"


class MachineJSONEncoder(json.JSONEncoder):
    """
    A JSONEncoder for machine objects, handling datetime.datetime objects as strings
    and :class:`~gordo_core.sensor_tag.SensorTag` as a dict
    representation of a :class:`~gordo.machine.Machine`
    """

    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.strftime(DATETIME_FORMAT)
        elif isinstance(obj, SensorTag):
            return obj.to_json()
        return super().default(obj)


class multiline_str(str):
    pass


class MachineSafeDumper(yaml.SafeDumper):
    pass


def _multiline_str_representer(dumper: yaml.SafeDumper, data: Any):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


def _datetime_representer(dumper: yaml.SafeDumper, data: Any):
    return dumper.represent_str(data.strftime(DATETIME_FORMAT))


def _sensor_tag_representer(dumper: yaml.SafeDumper, data: Any):
    return dumper.represent_dict(data.to_json())


MachineSafeDumper.add_representer(multiline_str, _multiline_str_representer)
MachineSafeDumper.add_representer(datetime, _datetime_representer)
MachineSafeDumper.add_representer(SensorTag, _sensor_tag_representer)



from typing import get_origin, get_args, Union

try:
    from types import UnionType  # type: ignore
except ImportError:
    UnionType = None


def _is_exact_tuple_type(tp) -> bool:
    if tp is tuple:
        return True
    origin = get_origin(tp)
    return origin is tuple


def is_tuple_type(tp) -> bool:
    """
    Check if this type is a tuple.

    Examples
    --------
    >>> from typing import Optional, Tuple
    >>> is_tuple_type(tuple)
    True
    >>> is_tuple_type(Optional[tuple[int, int]])
    True
    >>> is_tuple_type(Tuple[str, str])
    True
    >>> is_tuple_type(list[str])
    False

    Parameters
    ----------
    tp
        Type for check.

    Returns
    -------
    """
    if _is_exact_tuple_type(tp):
        return True
    origin = get_origin(tp)
    if origin is Union or (UnionType is not None and origin is UnionType):
        args = get_args(tp)
        for arg in args:
            if not _is_exact_tuple_type(arg) and not (arg is type(None)):
                return False
        return True
    return False



import pytest
import peewee
import json
import yaml

from gordo.machine import Machine
from gordo.machine.constants import MACHINE_YAML_FIELDS
from gordo.reporters.postgres import PostgresReporter, Machine as PostgresMachine
from gordo_core.sensor_tag import SensorTag


def test_builder_with_reporter(postgresdb, metadata):
    """
    Verify a model can take a reporter and .report() will run any given reporters
    """
    reporter = PostgresReporter(host="localhost")
    metadata["runtime"]["reporters"].append(reporter.to_dict())

    machine = Machine.from_dict(metadata)

    with pytest.raises(peewee.DoesNotExist):
        PostgresMachine.get(PostgresMachine.name == machine.name)
    machine.report()
    PostgresMachine.get(PostgresMachine.name == machine.name)


def create_machine():
    return Machine.from_config(
        dict(
            name="special-model-name",
            model={"sklearn.decomposition.PCA": {"svd_solver": "auto"}},
            dataset={
                "type": "RandomDataset",
                "train_start_date": "2017-12-25 06:00:00Z",
                "train_end_date": "2017-12-30 06:00:00Z",
                "tag_list": [SensorTag("Tag 1"), SensorTag("Tag 2")],
                "target_tag_list": [SensorTag("Tag 3"), SensorTag("Tag 4")],
            },
            project_name="test-proj",
        )
    )


def test_to_json():
    machine = create_machine()
    json_result = machine.to_json()
    result = json.loads(json_result)
    for field in MACHINE_YAML_FIELDS:
        if field in result:
            json.loads(result[field])


def test_to_yaml():
    machine = create_machine()
    yaml_result = machine.to_yaml()
    result = yaml.load(yaml_result)
    for field in MACHINE_YAML_FIELDS:
        if field in result:
            yaml.load(result[field])



import pytest

from gordo.workflow.config_elements.normalized_config import NormalizedConfig


def test_splited_docker_images():
    config = {"machines": [], "globals": {"runtime": {}}}
    normalized_config = NormalizedConfig(config, "test", "1.0.0")
    config_globals = normalized_config.globals
    config_runtime = config_globals["runtime"]
    assert config_runtime["deployer"]["image"] == "gordo-deploy"
    assert config_runtime["server"]["image"] == "gordo-model-server"
    assert config_runtime["prometheus_metrics_server"]["image"] == "gordo-model-server"
    assert config_runtime["builder"]["image"] == "gordo-model-builder"
    assert config_runtime["client"]["image"] == "gordo-client"


def test_unified_docker_images():
    config = {"machines": [], "globals": {"runtime": {}}}
    normalized_config = NormalizedConfig(config, "test", "1.3.0")
    config_globals = normalized_config.globals
    config_runtime = config_globals["runtime"]
    assert config_runtime["deployer"]["image"] == "gordo-base"
    assert config_runtime["server"]["image"] == "gordo-base"
    assert config_runtime["prometheus_metrics_server"]["image"] == "gordo-base"
    assert config_runtime["builder"]["image"] == "gordo-base"
    assert config_runtime["client"]["image"] == "gordo-base"


def test_custom_docker_images():
    config = {
        "machines": [],
        "globals": {
            "runtime": {
                "deployer": {"image": "my-deployer"},
                "server": {"image": "my-server"},
                "builder": {"image": "my-builder"},
            }
        },
    }
    normalized_config = NormalizedConfig(config, "test", "1.1.0")
    config_globals = normalized_config.globals
    config_runtime = config_globals["runtime"]
    assert config_runtime["deployer"]["image"] == "my-deployer"
    assert config_runtime["server"]["image"] == "my-server"
    assert config_runtime["prometheus_metrics_server"]["image"] == "gordo-model-server"
    assert config_runtime["builder"]["image"] == "my-builder"
    assert config_runtime["client"]["image"] == "gordo-client"



import pytest

from gordo.util.version import (
    parse_version,
    GordoRelease,
    GordoSpecial,
    Special,
    GordoPR,
    GordoSHA,
)


def test_release():
    release = GordoRelease(1, 2, 3)
    assert not release.only_major()
    assert not release.only_major_minor()
    assert not release.without_patch()
    release = GordoRelease(1, 2)
    assert not release.only_major()
    assert release.only_major_minor()
    assert release.without_patch()
    release = GordoRelease(1)
    assert release.only_major()
    assert not release.only_major_minor()
    assert release.without_patch()


@pytest.mark.parametrize(
    "gordo_version,expected",
    [
        ("1.2.3", GordoRelease(1, 2, 3)),
        ("3.4.5dev2", GordoRelease(3, 4, 5, "dev2")),
        ("0.1.dev0", GordoRelease(0, 1, None, ".dev0")),
        ("5.7", GordoRelease(5, 7)),
        ("latest", GordoSpecial(Special.LATEST)),
        ("pr-43", GordoPR(43)),
        ("dke0832k", GordoSHA("dke0832k")),
    ],
)
def test_versions(gordo_version, expected):
    result = parse_version(gordo_version)
    assert result == expected
    assert result.get_version() == gordo_version


def test_exceptions():
    with pytest.raises(ValueError):
        parse_version("pr-2d")
    with pytest.raises(ValueError):
        parse_version("0-1")



# -*- coding: utf-8 -*-
"""
This module contains code for generating the Gordo server Flask application.

Running this module will run the application using Flask's development webserver.
Gunicorn can be used to run the application as `gevent` async workers by using the
:func:`~gordo.server.server.run_server` function.
"""
import os
import json
import logging
import timeit
import typing
import subprocess
from functools import wraps

import yaml
from flask import Flask, g, request, current_app, make_response, jsonify

from typing import Optional, Any, Dict

from gordo.server import blueprints
from gordo import __version__

from prometheus_client import CollectorRegistry
from .prometheus import GordoServerPrometheusMetrics
from .utils import validate_revision

logger = logging.getLogger(__name__)


def enable_prometheus():
    return os.getenv("ENABLE_PROMETHEUS", "false") != "false"


class Config:
    """Server config"""

    def __init__(self):
        self.MODEL_COLLECTION_DIR_ENV_VAR = "MODEL_COLLECTION_DIR"
        self.EXPECTED_MODELS = yaml.safe_load(os.getenv("EXPECTED_MODELS", "[]"))
        self.ENABLE_PROMETHEUS = enable_prometheus()
        self.PROJECT = os.getenv("PROJECT")


def adapt_proxy_deployment(wsgi_app: typing.Callable) -> typing.Callable:
    """
    Decorator specific to fixing behind-proxy-issues when on Kubernetes and
    using Envoy proxy.

    Parameters
    ----------
    wsgi_app: typing.Callable
        The underlying WSGI application of a flask app, for example

    Notes
    -----
    Special note about deploying behind Ambassador, or prefixed proxy paths in general:

    When deployed on kubernetes/ambassador there is a prefix in-front of the
    server. ie::

        /gordo/v0/some-project-name/some-target

    The server itself only knows about routes to the right of such a prefix:
    such as ``/metadata`` or ``/predictions`` when in reality, the full path is::

        /gordo/v0/some-project-name/some-target/metadata

    This is solved by getting the current application's assigned prefix,
    where ``HTTP_X_ENVOY_ORIGINAL_PATH`` is the *full* path, including the prefix.
    and ``PATH_INFO`` is the actual relative path the server knows about.

    This function wraps the WSGI app itself to map the current full path
    to the assigned route function.

    ie. ``/metadata`` -> metadata route function, by default, but updates
    ``/gordo/v0/some-project-name/some-target/metadata`` -> metadata route function

    Returns
    -------
    Callable

    Example
    -------
    >>> app = Flask(__name__)
    >>> app.wsgi_app = adapt_proxy_deployment(app.wsgi_app)
    """

    @wraps(wsgi_app)
    def wrapper(environ, start_response):

        # Script name can be "/gordo/v0/some-project-name/some-target/metadata"
        script_name = environ.get("HTTP_X_ENVOY_ORIGINAL_PATH", "")
        if script_name:

            # PATH_INFO could be either "/" or some local route such as "/metadata"
            path_info = environ.get("PATH_INFO", "")
            if path_info.rstrip("/"):

                # PATH_INFO must be something like "/metadata" or other local path
                # To figure out the prefix/script_name we remove it from the
                # full HTTP_X_ENVOY_ORIGINAL_PATH, so that something such as
                # /gordo/v0/some-project-name/some-target/metadata, becomes
                # /gordo/v0/some-project-name/some-target/
                script_name = script_name.replace(path_info, "")
            environ["SCRIPT_NAME"] = script_name

            # Now we can just ensure the PATH_INFO reflects the locally known path
            # such as /metadata and not /gordo/v0/some-project-name/some-target/metadata
            if path_info.startswith(script_name):
                environ["PATH_INFO"] = path_info[len(script_name) :]

        scheme = environ.get("HTTP_X_FORWARDED_PROTO", "")
        if scheme:
            environ["wsgi.url_scheme"] = scheme
        return wsgi_app(environ, start_response)

    return wrapper


def create_prometheus_metrics(
    project: Optional[str] = None, registry: Optional[CollectorRegistry] = None
) -> GordoServerPrometheusMetrics:
    arg_labels = [("gordo_name", "model")]
    info = {"version": __version__}
    if project is not None:
        info["project"] = project
    else:
        arg_labels.append(("gordo_project", "project"))
    return GordoServerPrometheusMetrics(
        args_labels=arg_labels,
        info=info,
        ignore_paths=["/healthcheck"],
        registry=registry,
    )


def build_app(
    config: Optional[Dict[str, Any]] = None,
    prometheus_registry: Optional[CollectorRegistry] = None,
):
    """
    Build app and any associated routes
    """
    logger.info("Starting Flask('%s') server", __name__)
    app = Flask(__name__)
    app.config.from_object(Config())
    if config is not None:
        app.config.update(**config)

    app.register_blueprint(blueprints.base_blueprint)
    app.register_blueprint(blueprints.anomaly_blueprint)

    app.wsgi_app = adapt_proxy_deployment(app.wsgi_app)  # type: ignore
    app.url_map.strict_slashes = False  # /path and /path/ are ok.

    if app.config["ENABLE_PROMETHEUS"]:
        prometheus_metrics = create_prometheus_metrics(
            project=app.config.get("PROJECT"), registry=prometheus_registry
        )
        prometheus_metrics.prepare_app(app)
    elif prometheus_registry is not None:
        logger.warning("Ignoring non empty prometheus_registry argument")

    @app.before_request
    def _start_timer():
        g.start_time = timeit.default_timer()

    @app.before_request
    def _set_revision_and_collection_dir():
        g.collection_dir = os.environ[
            current_app.config["MODEL_COLLECTION_DIR_ENV_VAR"]
        ]
        g.current_revision = os.path.basename(g.collection_dir)

        # If a specific revision was requested, update collection_dir
        has_revision = "revision" in request.args or "revision" in request.headers
        if has_revision:
            if "revision" in request.args:
                g.revision = request.args["revision"]
            else:
                g.revision = request.headers["revision"]
            if not validate_revision(g.revision):
                return make_response(
                    jsonify({"error": "Revision should only contains numbers."}), 410
                )
            g.collection_dir = os.path.join(g.collection_dir, "..", g.revision)
            try:
                os.listdir(g.collection_dir)  # List dir to ensure it exists
            except FileNotFoundError:
                return make_response(
                    jsonify({"error": f"Revision '{g.revision}' not found."}), 410
                )
        else:
            g.revision = g.current_revision

    @app.after_request
    def _revision_used(response):
        if response.is_json:
            data = response.get_json()
            if data is not None:
                data["revision"] = g.revision
                response.set_data(json.dumps(data).encode())
        response.headers["revision"] = g.revision
        return response

    @app.after_request
    def _log_time_taken(response):
        runtime_s = timeit.default_timer() - g.start_time
        logger.debug(f"Total runtime for request: {runtime_s}s")
        response.headers["Server-Timing"] = f"request_walltime_s;dur={runtime_s}"
        return response

    @app.route("/healthcheck")
    def base_healthcheck():
        return "", 200

    @app.route("/server-version")
    def server_version():
        return jsonify({"version": __version__})

    return app


def run_cmd(cmd):
    """
    Run a shell command and handle CalledProcessError and OSError types

    Note
    ----
    This function is abstracted from :func:`~gordo.server.server.run_server`
    in order to test the calling of commands that would allow the subprocess call to
    break, depending on how it is parameterized. For example, calling this without
    sending `stderr` to stdout will cause a segmentation fault when calling an
    executable that does not exist.
    """
    subprocess.check_call(cmd, stderr=subprocess.STDOUT)


def run_server(
    host: str,
    port: int,
    workers: int,
    log_level: str,
    config_module: Optional[str] = None,
    worker_connections: Optional[int] = None,
    threads: Optional[int] = None,
    worker_class: str = "gthread",
    server_app: str = "gordo.server.server:build_app()",
):
    """
    Run application with Gunicorn server using Gevent Async workers

    Parameters
    ----------
    host: str
        The host to run the server on.
    port: int
        The port to run the server on.
    workers: int
        The number of worker processes for handling requests.
    log_level: str
        The log level for the `gunicorn` webserver. Valid log level names can be found
        in the [gunicorn documentation](http://docs.gunicorn.org/en/stable/settings.html#loglevel).
    config_module: str
        The config module. Will be passed with `python:` [prefix](https://docs.gunicorn.org/en/stable/settings.html#config).
    worker_connections: int
        The maximum number of simultaneous clients per worker process.
    threads: str
        The number of worker threads for handling requests.
    worker_class: str
        The type of workers to use.
    server_app: str
        The application to run
    """

    cmd = [
        "gunicorn",
        "--bind",
        f"{host}:{port}",
        "--log-level",
        log_level,
        "--error-logfile",
        "-",
        "--access-logfile",
        "-",
        "--worker-class",
        worker_class,
        "--worker-tmp-dir",
        "/dev/shm",
        "--workers",
        str(workers),
    ]
    if config_module is not None:
        cmd.extend(("--config", "python:" + config_module))
    if worker_class == "gthread":
        if threads is not None:
            cmd.extend(("--threads", str(threads)))
    else:
        if worker_connections is not None:
            cmd.extend(("--worker-connections", str(worker_connections)))

    cmd.append(server_app)
    run_cmd(cmd)


if __name__ == "__main__":
    app = build_app()

    # Run development webserver
    app.run("0.0.0.0", 5555, debug=True, threaded=False)



# -*- coding: utf-8 -*-

import io
from typing import Iterable, Tuple, Union

from sklearn.base import BaseEstimator

from gordo.workflow.config_elements.normalized_config import NormalizedConfig
from gordo.workflow.workflow_generator.workflow_generator import get_dict_from_yaml
from gordo.builder import ModelBuilder
from gordo.machine import Machine


def local_build(
    config_str: str,
) -> Iterable[Tuple[Union[BaseEstimator, None], Machine]]:
    """
    Build model(s) from a bare Gordo config file locally.

    This is very similar to the same steps as the normal workflow generation and subsequent
    Gordo deployment process makes. Should help developing locally,
    as well as giving a good indication that your config is valid for deployment
    with Gordo.

    Parameters
    ----------
    config_str: str
        The raw yaml config file in string format.

    Examples
    --------
    >>> import numpy as np
    >>> config = '''
    ... machines:
    ...       - dataset: |
    ...           tags:
    ...             - SOME-TAG1
    ...             - SOME-TAG2
    ...           target_tag_list:
    ...             - SOME-TAG3
    ...             - SOME-TAG4
    ...           train_end_date: '2019-03-01T00:00:00+00:00'
    ...           train_start_date: '2019-01-01T00:00:00+00:00'
    ...           asset: asgb
    ...           data_provider:
    ...             type: RandomDataProvider
    ...         metadata: |
    ...           information: Some sweet information about the model
    ...         model: |
    ...           gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
    ...             base_estimator:
    ...               sklearn.pipeline.Pipeline:
    ...                 steps:
    ...                 - sklearn.decomposition.PCA
    ...                 - sklearn.multioutput.MultiOutputRegressor:
    ...                     estimator: sklearn.linear_model.LinearRegression
    ...         name: crazy-sweet-name
    ... '''
    >>> models_n_metadata = local_build(config)
    >>> assert len(list(models_n_metadata)) == 1

    Returns
    -------
    Iterable[Tuple[Union[BaseEstimator, None], Machine]]
        A generator yielding tuples of models and their metadata.
    """

    config = get_dict_from_yaml(io.StringIO(config_str))
    normed = NormalizedConfig(config, project_name="local-build")
    for machine in normed.machines:
        yield ModelBuilder(machine=machine).build()



# -*- coding: utf-8 -*-

import datetime
import hashlib
import json
import pydoc
import logging
import os
import time
import random
from pathlib import Path
from typing import Union, Optional, Dict, Any, Tuple, Type, List, Callable

import pandas as pd
import numpy as np
import tensorflow as tf
import xarray as xr

import sklearn
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn import metrics
from sklearn.model_selection import BaseCrossValidator, cross_validate
from sklearn.pipeline import Pipeline

from gordo.util import disk_registry
from gordo import (
    serializer,
    __version__,
    parse_version,
)
from gordo_core.base import GordoBaseDataset
from gordo_core.import_utils import BackCompatibleLocations
from gordo.machine.model.base import GordoBase
from gordo.machine.model.utils import metric_wrapper
from gordo.workflow.config_elements.normalized_config import NormalizedConfig
from gordo.machine import Machine, load_model_config
from gordo.machine.metadata import (
    BuildMetadata,
    ModelBuildMetadata,
    DatasetBuildMetadata,
    CrossValidationMetaData,
)


logger = logging.getLogger(__name__)


class ModelBuilder:
    def __init__(
        self,
        machine: Machine,
        back_compatibles: Optional[BackCompatibleLocations] = None,
        default_data_provider: Optional[str] = None,
    ):
        """
        Build a model for a given :class:`gordo.machine.Machine`

        Parameters
        ----------
        machine: Machine

        Example
        -------
        >>> from gordo_core.sensor_tag import SensorTag
        >>> from gordo.machine import Machine
        >>> machine = Machine.from_config(dict(
        ...     name="special-model-name",
        ...     model={"sklearn.decomposition.PCA": {"svd_solver": "auto"}},
        ...     dataset={
        ...         "type": "RandomDataset",
        ...         "train_start_date": "2017-12-25 06:00:00Z",
        ...         "train_end_date": "2017-12-30 06:00:00Z",
        ...         "tag_list": [SensorTag("Tag 1"), SensorTag("Tag 2")],
        ...         "target_tag_list": [SensorTag("Tag 3"), SensorTag("Tag 4")]
        ...     },
        ...     project_name='test-proj',
        ... ))
        >>> builder = ModelBuilder(machine=machine)
        >>> model, machine = builder.build()
        """
        # Avoid overwriting the passed machine, copy doesn't work if it holds
        # reference to a loaded Tensorflow model; .to_dict() serializes it to
        # a primitive dict representation.
        self.machine = Machine.from_dict(
            machine.to_dict(),
            back_compatibles=back_compatibles,
            default_data_provider=default_data_provider,
        )
        self.back_compatibles = back_compatibles
        self.default_data_provider = default_data_provider

    @property
    def cached_model_path(self) -> Union[os.PathLike, str, None]:
        return getattr(self, "_cached_model_path", None)

    @cached_model_path.setter
    def cached_model_path(self, value):
        self._cached_model_path = value

    @property
    def gordo_version(self):
        return __version__

    def build(
        self,
        output_dir: Optional[Union[os.PathLike, str]] = None,
        model_register_dir: Optional[Union[os.PathLike, str]] = None,
        replace_cache=False,
    ) -> Tuple[sklearn.base.BaseEstimator, Machine]:
        """
        Always return a model and its metadata.

        If ``output_dir`` is supplied, it will save the model there.
        ``model_register_dir`` points to the model cache directory which it will
        attempt to read the model from. Supplying both will then have the effect
        of both; reading from the cache and saving that cached model to the new
        output directory.

        Parameters
        ----------
        output_dir: Optional[Union[os.PathLike, str]]
            A path to where the model will be deposited.
        model_register_dir: Optional[Union[os.PathLike, str]]
            A path to a register, see `:func:gordo.util.disk_registry`.
            If this is None then always build the model, otherwise try to resolve
            the model from the registry.
        replace_cache: bool
            Forces a rebuild of the model, and replaces the entry in the cache
            with the new model.

        Returns
        -------
        Tuple[sklearn.base.BaseEstimator, Machine]
            Built model and an updated ``Machine``
        """
        if not model_register_dir:
            model, machine = self._build()
        else:
            cache_key = self.cache_key
            logger.debug(
                f"Model caching activated, attempting to read model-location with key "
                f"{cache_key} from register {model_register_dir}"
            )
            self.cached_model_path = self.check_cache(model_register_dir, cache_key)

            if replace_cache:
                logger.info("replace_cache=True, deleting any existing cache entry")
                disk_registry.delete_value(model_register_dir, cache_key)
                self.cached_model_path = None

            # Load the model from previous cached directory
            if self.cached_model_path:
                model = serializer.load(self.cached_model_path)
                metadata = serializer.load_metadata(self.cached_model_path)
                metadata["metadata"][
                    "user_defined"
                ] = self.machine.metadata.user_defined

                metadata["runtime"] = self.machine.runtime

                machine = Machine.from_dict(
                    load_model_config(metadata),
                    back_compatibles=self.back_compatibles,
                    default_data_provider=self.default_data_provider,
                )

            # Otherwise build and cache the model
            else:
                model, machine = self._build()
                cache_key = self.calculate_cache_key(machine)
                self.cached_model_path = self._save_model(
                    model=model,
                    machine=machine,
                    output_dir=output_dir,  # type: ignore
                    checksum=cache_key,
                )
                logger.info(
                    f"Built model, and deposited at {self.cached_model_path} with checksum '{cache_key}'"
                )
                logger.info("Writing model-location to model registry")
                disk_registry.write_key(  # type: ignore
                    model_register_dir, cache_key, self.cached_model_path
                )

        # Save model to disk, if we're not building for cv only purposes.
        if output_dir and (self.machine.evaluation.get("cv_mode") != "cross_val_only"):
            cache_key = self.calculate_cache_key(machine)
            self.cached_model_path = self._save_model(
                model=model, machine=machine, output_dir=output_dir, checksum=cache_key
            )
        return model, machine

    def _build(self) -> Tuple[sklearn.base.BaseEstimator, Machine]:
        """
        Build the model using the current state of the Builder

        Returns
        -------
            Tuple[sklearn.base.BaseEstimator, dict]
        """
        # Enforce random seed to 0 if not specified.
        self.set_seed(seed=self.machine.evaluation.get("seed", 0))

        # Get the dataset from config
        logger.debug(
            f"Initializing Dataset with config {self.machine.dataset.to_dict()}"
        )

        dataset = GordoBaseDataset.from_dict(self.machine.dataset.to_dict())

        logger.debug("Fetching training data")
        start = time.time()

        X, y = dataset.get_data()

        time_elapsed_data = time.time() - start

        # Get the model and dataset
        logger.debug(f"Initializing Model with config: {self.machine.model}")
        model = serializer.from_definition(self.machine.model)

        cv_duration_sec = None

        machine: Machine = Machine.from_dict(
            dict(
                name=self.machine.name,
                dataset=self.machine.dataset.to_dict(),
                metadata=self.machine.metadata.to_dict(),
                model=self.machine.model,
                project_name=self.machine.project_name,
                evaluation=self.machine.evaluation,
                runtime=self.machine.runtime,
            ),
            back_compatibles=self.back_compatibles,
            default_data_provider=self.default_data_provider,
        )

        split_metadata: Dict[str, Any] = dict()
        scores: Dict[str, Any] = dict()
        if self.machine.evaluation["cv_mode"].lower() in (
            "cross_val_only",
            "full_build",
        ):

            # Build up a metrics list.
            metrics_list = self.metrics_from_list(
                self.machine.evaluation.get("metrics")
            )

            # Cross validate
            if hasattr(model, "predict"):
                logger.debug("Starting cross validation")
                start = time.time()

                scaler = self.machine.evaluation.get("scoring_scaler")
                metrics_dict = self.build_metrics_dict(metrics_list, y, scaler=scaler)

                split_obj = serializer.from_definition(
                    self.machine.evaluation.get(
                        "cv",
                        {"sklearn.model_selection.TimeSeriesSplit": {"n_splits": 3}},
                    )
                )
                # Generate metadata about CV train, test splits
                split_metadata = ModelBuilder.build_split_dict(X, split_obj)

                cv_kwargs = dict(
                    X=X, y=y, scoring=metrics_dict, return_estimator=True, cv=split_obj
                )
                if hasattr(model, "cross_validate"):
                    cv = model.cross_validate(**cv_kwargs)
                else:
                    cv = cross_validate(model, **cv_kwargs)

                for metric, test_metric in map(
                    lambda k: (k, f"test_{k}"), metrics_dict
                ):
                    val = {
                        "fold-mean": cv[test_metric].mean(),
                        "fold-std": cv[test_metric].std(),
                        "fold-max": cv[test_metric].max(),
                        "fold-min": cv[test_metric].min(),
                    }
                    val.update(
                        {
                            f"fold-{i + 1}": raw_value
                            for i, raw_value in enumerate(cv[test_metric].tolist())
                        }
                    )
                    scores.update({metric: val})

                cv_duration_sec = time.time() - start
            else:
                logger.debug("Unable to score model, has no attribute 'predict'.")

            # If cross_val_only, return without fitting to the whole dataset
            if self.machine.evaluation["cv_mode"] == "cross_val_only":
                machine.metadata.build_metadata = BuildMetadata(
                    model=ModelBuildMetadata(
                        cross_validation=CrossValidationMetaData(
                            cv_duration_sec=cv_duration_sec,
                            scores=scores,
                            splits=split_metadata,
                        )
                    ),
                    dataset=DatasetBuildMetadata(
                        query_duration_sec=time_elapsed_data,
                        dataset_meta=dataset.get_metadata(),
                    ),
                )
                return model, machine

        # Train
        logger.debug("Starting to train model.")
        start = time.time()
        model.fit(X, y)
        time_elapsed_model = time.time() - start

        # Build specific metadata
        machine.metadata.build_metadata = BuildMetadata(
            model=ModelBuildMetadata(
                model_offset=self._determine_offset(model, X),
                model_creation_date=str(
                    datetime.datetime.now(datetime.timezone.utc).astimezone()
                ),
                model_builder_version=self.gordo_version,
                model_training_duration_sec=time_elapsed_model,
                cross_validation=CrossValidationMetaData(
                    cv_duration_sec=cv_duration_sec,
                    scores=scores,
                    splits=split_metadata,
                ),
                model_meta=self._extract_metadata_from_model(model),
            ),
            dataset=DatasetBuildMetadata(
                query_duration_sec=time_elapsed_data,
                dataset_meta=dataset.get_metadata(),
            ),
        )
        return model, machine

    def set_seed(self, seed: int):
        logger.info(f"Setting random seed: '{seed}'")
        tf.random.set_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    @staticmethod
    def build_split_dict(X: pd.DataFrame, split_obj: Type[BaseCrossValidator]) -> dict:
        """
        Get dictionary of cross-validation training dataset split metadata

        Parameters
        ----------
        X: pd.DataFrame
            The training dataset that will be split during cross-validation.
        split_obj: Type[sklearn.model_selection.BaseCrossValidator]
            The cross-validation object that returns train, test indices for splitting.

        Returns
        -------
        split_metadata: Dict[str,Any]
            Dictionary of cross-validation train/test split metadata
        """
        split_metadata: Dict[str, Any] = dict()
        for i, (train_ind, test_ind) in enumerate(split_obj.split(X)):
            split_metadata.update(
                {
                    f"fold-{i+1}-train-start": X.index[train_ind[0]],
                    f"fold-{i+1}-train-end": X.index[train_ind[-1]],
                    f"fold-{i+1}-test-start": X.index[test_ind[0]],
                    f"fold-{i+1}-test-end": X.index[test_ind[-1]],
                }
            )
            split_metadata.update({f"fold-{i+1}-n-train": len(train_ind)})
            split_metadata.update({f"fold-{i+1}-n-test": len(test_ind)})
        return split_metadata

    @staticmethod
    def build_metrics_dict(
        metrics_list: list,
        y: pd.DataFrame,
        scaler: Optional[Union[TransformerMixin, str]] = None,
    ) -> dict:
        """
        Given a list of metrics that accept a true_y and pred_y as inputs this returns a
        dictionary with keys in the form '{score}-{tag_name}' for each given target tag
        and '{score}' for the average score across all target tags and folds,
        and values being the callable make_scorer(metric_wrapper(score)). Note: score in
        {score}-{tag_name} is a sklearn's score function name with '_' replaced by '-'
        and tag_name corresponds to given target tag name with ' ' replaced by '-'.

        Parameters
        ----------
        metrics_list: list
            List of sklearn score functions
        y: pd.DataFrame
            Target data
        scaler : Optional[Union[TransformerMixin, str]]
            Scaler which will be fitted on y, and used to transform the data before
            scoring. Useful when the metrics are sensitive to the amplitude of the data, and
            you have multiple targets.


        Returns
        -------
            dict
        """
        if scaler:
            if isinstance(scaler, str) or isinstance(scaler, dict):
                scaler = serializer.from_definition(scaler)
            logger.debug("Fitting scaler for scoring purpose")
            scaler.fit(y)

        def _score_factory(metric_func=metrics.r2_score, col_index=0):
            def _score_per_tag(y_true, y_pred):
                # This function extracts the score for each given target_tag to
                # use as scoring argument in sklearn cross_validate, as the scoring
                # must return a single value.
                if hasattr(y_true, "values"):
                    y_true = y_true.values
                if hasattr(y_pred, "values"):
                    y_pred = y_pred.values

                return metric_func(y_true[:, col_index], y_pred[:, col_index])

            return _score_per_tag

        metrics_dict = {}
        for metric in metrics_list:
            for index, col in enumerate(y.columns):
                metric_str = metric.__name__.replace("_", "-")
                metrics_dict.update(
                    {
                        metric_str
                        + f'-{col.replace(" ", "-")}': metrics.make_scorer(
                            metric_wrapper(
                                _score_factory(metric_func=metric, col_index=index),
                                scaler=scaler,
                            )
                        )
                    }
                )

            metrics_dict.update(
                {metric_str: metrics.make_scorer(metric_wrapper(metric, scaler=scaler))}
            )
        return metrics_dict

    @staticmethod
    def _determine_offset(
        model: BaseEstimator, X: Union[np.ndarray, pd.DataFrame]
    ) -> int:
        """
        Determine the model's offset. How much does the output of the model differ
        from its input?

        Parameters
        ----------
        model: sklearn.base.BaseEstimator
            Trained model with either ``predict`` or ``transform`` method, preference
            given to ``predict``.
        X: Union[np.ndarray, pd.DataFrame]
            Data to pass to the model's ``predict`` or ``transform`` method.

        Returns
        -------
        int
            The difference between X and the model's output lengths.
        """
        if isinstance(X, pd.DataFrame) or isinstance(X, xr.DataArray):
            X = X.values
        out = model.predict(X) if hasattr(model, "predict") else model.transform(X)
        return len(X) - len(out)

    @staticmethod
    def _extract_metadata(machine: Union[Machine, dict]):
        if isinstance(machine, Machine):
            return machine.to_dict()
        else:
            return machine

    @staticmethod
    def _save_model(
        model: BaseEstimator,
        machine: Union[Machine, dict],
        output_dir: Union[os.PathLike, str],
        checksum: Optional[str] = None,
    ):
        """
        Save the model according to the expected Argo workflow procedure.
        Parameters
        ----------
        model: BaseEstimator
            The model to save to the directory with gordo serializer.
        machine: Union[Machine, dict]
            Machine instance used to build this model.
        output_dir: Union[os.PathLike, str]
            The directory where to save the model, will create directories if needed.
        checksum: Optional[str]
            Model revision sha512 checksum. Might be taken from `self.check key`
        Returns
        -------
        Union[os.PathLike, str]
            Path to the saved model
        """
        os.makedirs(output_dir, exist_ok=True)  # Ok if some dirs exist
        info: Optional[dict] = None
        if checksum is not None:
            info = {"checksum": checksum}
        serializer.dump(
            model,
            output_dir,
            metadata=machine.to_dict() if isinstance(machine, Machine) else machine,
            info=info,
        )
        return output_dir

    @staticmethod
    def _extract_metadata_from_model(
        model: BaseEstimator, metadata: dict = dict()
    ) -> dict:
        """
        Recursively check for :class:`gordo.machine.model.base.GordoBase` in a
        given ``model``. If such the model exists buried inside of a
        :class:`sklearn.pipeline.Pipeline` which is then part of another
        :class:`sklearn.base.BaseEstimator`, this function will return its metadata.

        Parameters
        ----------
        model: BaseEstimator
        metadata: dict
            Any initial starting metadata, but is mainly meant to be used during
            the recursive calls to accumulate any multiple
            :class:`gordo.machine.model.base.GordoBase` models found in this model

        Notes
        -----
        If there is a ``GordoBase`` model inside of a ``Pipeline`` which is not the final
        step, this function will not find it.

        Returns
        -------
        dict
            Dictionary representing accumulated calls to
            :meth:`gordo.machine.model.base.GordoBase.get_metadata`
        """
        metadata = metadata.copy()

        # If it's a Pipeline, only need to get the last step, which potentially has metadata
        if isinstance(model, Pipeline):
            final_step = model.steps[-1][1]
            metadata.update(ModelBuilder._extract_metadata_from_model(final_step))
            return metadata

        # GordoBase is simple, having a .get_metadata()
        if isinstance(model, GordoBase):
            metadata.update(model.get_metadata())

        # Continue to look at object values in case, we decided to have a GordoBase
        # which also had a GordoBase as a parameter/attribute, but will satisfy BaseEstimators
        # which can take a GordoBase model as a parameter, which will then have metadata to get
        for val in model.__dict__.values():
            if isinstance(val, Pipeline):
                metadata.update(
                    ModelBuilder._extract_metadata_from_model(val.steps[-1][1])
                )
            elif isinstance(val, GordoBase) or isinstance(val, BaseEstimator):
                metadata.update(ModelBuilder._extract_metadata_from_model(val))
        return metadata

    @property
    def cache_key(self) -> str:
        return self.calculate_cache_key(self.machine)

    def calculate_cache_key(self, machine: Machine) -> str:
        """
        Calculates a hash-key from the model and data-config.

        Returns
        -------
        str:
            A 512 byte hex value as a string based on the content of the parameters.

        Examples
        -------
        >>> from gordo.machine import Machine
        >>> from gordo_core.sensor_tag import SensorTag
        >>> machine = Machine.from_config(dict(
        ...     name="special-model-name",
        ...     model={"sklearn.decomposition.PCA": {"svd_solver": "auto"}},
        ...     dataset={
        ...         "type": "RandomDataset",
        ...         "train_start_date": "2017-12-25 06:00:00Z",
        ...         "train_end_date": "2017-12-30 06:00:00Z",
        ...         "tag_list": [SensorTag("Tag 1"), SensorTag("Tag 2")],
        ...         "target_tag_list": [SensorTag("Tag 3"), SensorTag("Tag 4")]
        ...     },
        ...     project_name='test-proj'
        ... ))
        >>> builder = ModelBuilder(machine)
        >>> len(builder.cache_key)
        128
        """
        # Sets a lot of the parameters to json.dumps explicitly to ensure that we get
        # consistent hash-values even if json.dumps changes their default values
        # (and as such might generate different json which again gives different hash)
        major_version, minor_version, is_unstable_version = parse_version(
            self.gordo_version
        )
        gordo_version = self.gordo_version if is_unstable_version else ""
        json_rep = json.dumps(
            {
                "name": machine.name,
                "model_config": machine.model,
                "data_config": machine.dataset.to_dict(),
                "evaluation_config": machine.evaluation,
                "gordo-major-version": major_version,
                "gordo-minor-version": minor_version,
                "gordo_version": gordo_version,
            },
            sort_keys=True,
            default=str,
            skipkeys=False,
            ensure_ascii=True,
            check_circular=True,
            allow_nan=True,
            cls=None,
            indent=None,
            separators=None,
        )
        logger.debug(f"Calculating model hash key for model: {json_rep}")
        return hashlib.sha3_512(json_rep.encode("ascii")).hexdigest()

    @staticmethod
    def check_cache(model_register_dir: Union[os.PathLike, str], cache_key: str):
        """
        Checks if the model is cached, and returns its path if it exists.

        Parameters
        ----------
        model_register_dir: [os.PathLike, None]
            The register dir where the model lies.
        cache_key: str
            A 512 byte hex value as a string based on the content of the parameters.

         Returns
        -------
        Union[os.PathLike, None]:
            The path to the cached model, or None if it does not exist.
        """
        existing_model_location = disk_registry.get_value(model_register_dir, cache_key)

        # Check that the model is actually there
        if existing_model_location and Path(existing_model_location).exists():
            logger.debug(
                f"Found existing model at path {existing_model_location}, returning it"
            )
            return existing_model_location
        elif existing_model_location:
            logger.warning(
                f"Found that the model-path {existing_model_location} stored in the "
                f"registry did not exist."
            )
            return None
        else:
            logger.info(
                f"Did not find the model with key {cache_key} in the register at "
                f"{model_register_dir}."
            )
            return None

    @staticmethod
    def metrics_from_list(metric_list: Optional[List[str]] = None) -> List[Callable]:
        """
        Given a list of metric function paths. ie. sklearn.metrics.r2_score or
        simple function names which are expected to be in the ``sklearn.metrics`` module,
        this will return a list of those loaded functions.

        Parameters
        ----------
        metrics: Optional[List[str]]
            List of function paths to use as metrics for the model Defaults to
            those specified in :class:`gordo.workflow.config_components.NormalizedConfig`
            sklearn.metrics.explained_variance_score,
            sklearn.metrics.r2_score,
            sklearn.metrics.mean_squared_error,
            sklearn.metrics.mean_absolute_error

        Returns
        -------
        List[Callable]
            A list of the functions loaded

        Raises
        ------
        AttributeError:
           If the function cannot be loaded.
        """
        defaults = NormalizedConfig.DEFAULT_CONFIG_GLOBALS["evaluation"]["metrics"]
        funcs = list()
        for func_path in metric_list or defaults:
            func = pydoc.locate(func_path)
            if func is None:
                # Final attempt, load function from sklearn.metrics module.
                funcs.append(getattr(metrics, func_path))
            else:
                funcs.append(func)
        return funcs



# -*- coding: utf-8 -*-

import io
import logging

import pytest
import pandas as pd
import numpy as np

from gordo.server import utils as server_utils
from gordo.server.properties import get_target_tags
from gordo_core.sensor_tag import SensorTag
from flask import Flask, g


def test_empty_target_tag_list():
    app = Flask(__name__)
    with app.app_context():
        g.metadata = {
            "dataset": {"tag_list": [SensorTag("test"), "test1"]},
            "metadata": {
                "build_metadata": {
                    "dataset": {
                        "dataset_meta": {
                            "tag_loading_metadata": {
                                "tags": {
                                    "test": {"name": "test"},
                                    "test1": {"name": "test1"},
                                }
                            }
                        }
                    }
                }
            },
        }
        assert get_target_tags() == [
            SensorTag("test"),
            SensorTag("test1"),
        ]


@pytest.mark.parametrize(
    "data_size,to_dict_arg",
    [(10, None), (1, None), (10, "records"), (10, "list"), (10, "dict")],
)
@pytest.mark.parametrize("resp_format", ("json", "parquet", None))
@pytest.mark.parametrize("send_as_parquet", (True, False))
def test_prediction_endpoint_post_ok(
    base_route,
    sensors,
    sensors_str,
    gordo_ml_server_client,
    data_size,
    to_dict_arg,
    resp_format,
    send_as_parquet,
):
    """
    Test the expected successful data posts, by sending a variety of valid
    JSON formats of a dataframe, as well as parquet serializations.
    """
    data_to_post = np.random.random(size=(data_size, len(sensors))).tolist()

    if to_dict_arg is not None:
        df = pd.DataFrame(data_to_post, columns=sensors_str)
        data_to_post = df.to_dict(to_dict_arg)

    endpoint = f"{base_route}/prediction"
    if resp_format is not None:
        endpoint += f"?format={resp_format}"

    if send_as_parquet:
        X = pd.DataFrame.from_dict(data_to_post)
        kwargs = dict(
            data={"X": (io.BytesIO(server_utils.dataframe_into_parquet_bytes(X)), "X")}
        )
    else:
        kwargs = dict(json={"X": data_to_post})

    resp = gordo_ml_server_client.post(endpoint, **kwargs)
    assert resp.status_code == 200

    if resp_format in (None, "json"):
        data = server_utils.dataframe_from_dict(resp.json["data"])
    else:
        data = server_utils.dataframe_from_parquet_bytes(resp.data)

    # Expected column names
    assert all(key in data for key in ("model-output", "model-input"))


def test_prediction_endpoint_post_fail(
    caplog, base_route, sensors, gordo_ml_server_client
):
    """
    Test expected failures when posting certain types of data
    """
    erroneous_datasets = [
        # Not enough features
        {"X": [list(range(len(sensors) - 1))]},
        # No X
        {"no-X-here": True},
        # Badly formatted data
        {"X": [[1, 2, [1, 2]]]},
    ]
    for i, data_to_post in enumerate(erroneous_datasets):
        with caplog.at_level(logging.CRITICAL):
            resp = gordo_ml_server_client.post(
                f"{base_route}/prediction", json=data_to_post
            )
        assert (
            resp.status_code == 400
        ), f"Prediction did not fail with data, {str(data_to_post)}."



import datetime
import json

import mlflow
from mlflow.entities import Metric, Param
import mock
import pytest

import gordo.reporters.mlflow as mlu
from gordo.machine.encoders import MachineJSONEncoder
from gordo.machine import Machine
from gordo.reporters.exceptions import ReporterException


def test_validate_dict():
    """
    Test that error is raised when a key is missing from the passed dictionary
    """
    required_keys = ["a", "b", "c"]
    # All keys present passes
    assert mlu._validate_dict({"a": 1, "b": 2, "c": 3}, required_keys) is None
    # Key missing fails
    with pytest.raises(ReporterException):
        mlu._validate_dict({"a": 1, "b": 2}, required_keys)


@pytest.mark.parametrize(
    "workspace_kwargs,service_principal_kwargs,n_interactive,n_spauth,n_get_uri",
    [
        # Empty config, local MLflow
        ({}, {}, 0, 0, 0),
        # Config with no auth, Azure Interactive Login
        (
            {
                "subscription_id": "dummy",
                "resource_group": "dummy",
                "workspace_name": "dummy",
            },
            {},
            1,
            0,
            1,
        ),
        # Config with auth, Azure ServicePrincipal Auth
        (
            {
                "subscription_id": "dummy",
                "resource_group": "dummy",
                "workspace_name": "dummy",
            },
            {
                "tenant_id": "dummy",
                "service_principal_id": "dummy",
                "service_principal_password": "dummy",
            },
            0,
            1,
            1,
        ),
    ],
)
def test_get_mlflow_client(
    workspace_kwargs, service_principal_kwargs, n_interactive, n_spauth, n_get_uri
):
    """
    Test that external methods are called correctly given different configurations
    """

    with mock.patch("gordo.reporters.mlflow.Workspace") as MockWorkspace, mock.patch(
        "gordo.reporters.mlflow.InteractiveLoginAuthentication"
    ) as MockInteractiveAuth, mock.patch(
        "gordo.reporters.mlflow.ServicePrincipalAuthentication"
    ) as MockSPAuth, mock.patch(
        "gordo.reporters.mlflow.MlflowClient"
    ) as MockClient:
        MockInteractiveAuth.return_value = True
        MockWorkspace.return_value.get_mlflow_tracking_uri.return_value = "test_uri"
        mlu.get_mlflow_client(workspace_kwargs, service_principal_kwargs)
        assert MockInteractiveAuth.call_count == n_interactive
        assert MockSPAuth.call_count == n_spauth
        assert MockWorkspace.call_count == n_get_uri
        MockClient.assert_called_once()


@mock.patch("gordo.reporters.mlflow.InteractiveLoginAuthentication")
@mock.patch("gordo.reporters.mlflow.Workspace")
@mock.patch("gordo.reporters.mlflow.MlflowClient")
def test_get_mlflow_client_config(MockClient, MockWorkspace, MockInteractiveLogin):
    """
    Test that error is raised with incomplete kwargs
    """
    # Incomplete workspace kwargs
    with pytest.raises(ReporterException):
        mlu.get_mlflow_client(
            {"subscription_id": "dummy", "workspace_name": "dummy"}, {}
        )

    # Incomplete service principal kwargs
    with pytest.raises(ReporterException):
        mlu.get_mlflow_client(
            {
                "subscription_id": "dummy",
                "resource_group": "dummy",
                "workspace_name": "dummy",
            },
            {"tenant_id": "dummy", "service_principal_password": "dummy"},
        )


@mock.patch("gordo.reporters.mlflow.MlflowClient.get_experiment_by_name")
@mock.patch("gordo.reporters.mlflow.MlflowClient.create_experiment")
@mock.patch("gordo.reporters.mlflow.MlflowClient.create_run")
def test_get_run_id_external_calls(
    mock_create_run, mock_create_experiment, mock_get_experiment, tmpdir
):
    """
    Test logic for creating an experiment if it does not exist to create new runs
    """

    class MockRunInfo:
        def __init__(self, run_id):
            self.run_id = run_id

    class MockRun:
        def __init__(self, run_id):
            self.info = MockRunInfo(run_id)

    class MockExperiment:
        def __init__(self, experiment_id):
            self.experiment_id = experiment_id

    def _test_calls(test_run_id, n_create_exp, n_create_run):
        """Test that number of calls match those specified"""
        run_id = mlu.get_run_id(client, experiment_name, model_key)
        assert mock_get_experiment.call_count == 1
        assert mock_create_experiment.call_count == n_create_exp
        assert mock_create_run.call_count == n_create_run
        assert run_id == test_run_id

    # Dummy test name/IDs
    experiment_name = "test_experiment"
    test_experiment_id = "dummy_exp_id"
    test_run_id = "dummy_run_id"
    model_key = "dummy_model_key"

    mlflow.set_tracking_uri(f"file:{tmpdir}")
    client = mlu.MlflowClient()

    # Experiment exists
    # Create a run with existing experiment_id
    mock_get_experiment.return_value = MockExperiment(test_experiment_id)
    mock_create_experiment.return_value = MockExperiment(test_experiment_id)
    mock_create_run.return_value = MockRun(test_run_id)
    _test_calls(test_run_id, n_create_exp=0, n_create_run=1)

    # Reset call counts
    for m in [mock_get_experiment, mock_create_experiment, mock_create_run]:
        m.call_count = 0

    # Experiment doesn't exist
    # Create an experiment and use its ID to create a run
    mock_get_experiment.return_value = None
    mock_create_experiment.return_value = MockExperiment(test_experiment_id)
    mock_create_run.return_value = MockRun(test_run_id)
    _test_calls(test_run_id, n_create_exp=1, n_create_run=1)


@pytest.mark.parametrize(
    "x,expected",
    [
        (datetime.datetime(1970, 1, 1), 0),
        (datetime.datetime(1970, 1, 1, 0, 0, 1), 1000),
        (datetime.datetime(1970, 1, 1, 0, 0, 0, 1000), 1),
    ],
)
def test_datetime_to_ms_since_epoch(x, expected):
    """
    Test that datetime is correctly converted to ms since Unix epoch
    """
    assert mlu._datetime_to_ms_since_epoch(x) == expected


@pytest.mark.parametrize(
    "n_metrics,n_max_metrics,n_params,n_max_params,expected_n_batches",
    [
        # N max metrics limiting
        (20, 5, 20, 20, 4),
        # N max params limiting
        (20, 20, 20, 5, 4),
        # N max equal
        (20, 5, 20, 5, 4),
    ],
)
def test_batch_log_items(
    n_metrics, n_max_metrics, n_params, n_max_params, expected_n_batches
):
    metrics = ["dummy_metric"] * n_metrics
    params = ["dummy_param"] * n_params

    log_batches = mlu.batch_log_items(metrics, params, n_max_metrics, n_max_params)

    # Make sure that we get the number of batches based on max number of metrics, params allowed
    assert len(log_batches) == expected_n_batches

    # Regardless of batch, each batch kwargs should have the same keys, "metrics" and "params"
    assert all([("metrics" in b) and ("params" in b) for b in log_batches])


def test_get_machine_log_items(metadata):
    """
    Test that dicts are correctly converted to MLflow types or errors raised
    """
    metrics, params = mlu.get_machine_log_items(Machine.from_dict(metadata))

    assert all(type(m) == Metric for m in metrics)
    assert all(type(p) == Param for p in params)


@pytest.mark.parametrize(
    "secret_str,keys,keys_valid",
    [
        ("dummy1:dummy2:dummy3", ["key1", "key2", "key3"], True),
        ("dummy1:dummy2:dummy3", ["key1", "key2"], False),
    ],
)
def test_get_kwargs_from_secret(monkeypatch, secret_str, keys, keys_valid):
    """
    Test that service principal kwargs are generated correctly if env var present
    """
    env_var_name = "TEST_SECRET"

    # TEST_SECRET doesn't exist as env var
    with pytest.raises(ReporterException):
        mlu.get_kwargs_from_secret(env_var_name, keys)

    # TEST_SECRET exists as env var
    monkeypatch.setenv(name=env_var_name, value=secret_str)
    if keys_valid:
        kwargs = mlu.get_kwargs_from_secret(env_var_name, keys)
        for key, value in zip(keys, secret_str.split(":")):
            assert kwargs[key] == value
    else:
        with pytest.raises(ReporterException):
            mlu.get_kwargs_from_secret(env_var_name, keys)


def test_workspace_spauth_kwargs():
    """Make sure an error is thrown when env vars not set"""
    with pytest.raises(ReporterException):
        mlu.get_workspace_kwargs()

    with pytest.raises(ReporterException):
        mlu.get_spauth_kwargs()


def test_MachineEncoder(metadata):
    """
    Test that metadata can dump successfully with MachineEncoder
    """
    assert json.dumps(metadata, cls=MachineJSONEncoder)


@mock.patch("gordo.reporters.mlflow.MlflowClient", autospec=True)
def test_mlflow_context_log_metadata(MockClient, tmpdir, metadata):
    """
    Test that call to wrapped function initiates MLflow logging or throws warning
    """
    metadata = Machine.from_dict(metadata)
    mlflow.set_tracking_uri(f"file:{tmpdir}")

    mock_client = MockClient()
    mock_client.log_batch.return_value = "test"

    # Function with a metadata dict returned
    with mlu.mlflow_context("returns metadata", "unique_key", {}, {}) as (
        mlflow_client,
        run_id,
    ):
        mlu.log_machine(mlflow_client, run_id, metadata)

    assert mock_client.log_batch.called


@mock.patch("gordo.reporters.mlflow.MlflowClient", autospec=True)
def test_mlflow_context_log_error(MockClient, metadata):
    """
    Test that an error while logging metadata as an artifact raises MlflowLoggingError
    """
    metadata = Machine.from_dict(metadata)
    mock_client = MockClient()
    mock_client.log_artifacts.side_effect = Exception("Some unknown exception!")

    with pytest.raises(mlu.MlflowLoggingError):
        with mlu.mlflow_context("returns metadata", "unique_key", {}, {}) as (
            mlflow_client,
            run_id,
        ):
            mlu.log_machine(mlflow_client, run_id, metadata)



import copy

import pandas as pd

from typing import Optional, Any
from flask import g

from gordo_core.sensor_tag import SensorTag

from gordo.utils import normalize_sensor_tags


def find_path_in_dict(path: list[str], data: dict) -> Any:
    """
    Find a path in `dict` recursively

    Examples
    --------
    >>> find_path_in_dict(["parent", "child"], {"parent": {"child": 42}})
    42

    Parameters
    ----------
    path: List[str]
    data: dict

    Returns
    -------

    """
    reversed_path = copy.copy(path)
    reversed_path.reverse()
    curr_data = data
    while len(reversed_path):
        key = reversed_path.pop()
        if key not in curr_data:
            exception_path = ".".join(path[: len(path) - len(reversed_path)])
            raise KeyError("'%s' is absent" % exception_path)
        curr_data = curr_data[key]
    return curr_data


def get_frequency():
    """
    The frequency the model was trained with in the dataset
    """
    return pd.tseries.frequencies.to_offset(g.metadata["dataset"]["resolution"])


def load_build_dataset_metadata():
    try:
        build_dataset_metadata = find_path_in_dict(
            ["metadata", "build_metadata", "dataset"], g.metadata
        )
    except KeyError as e:
        raise ValueError("Unable to load build dataset metadata: %s" % str(e))
    return build_dataset_metadata


def get_normalize_additional_fields(dataset: dict[str, Any]):
    additional_fields: dict[str, Optional[str]] = {}
    if "default_tag" in dataset and dataset["default_tag"]:
        additional_fields = dataset["default_tag"]
    # Keeping back-compatibility for a while
    elif dataset.get("asset", None):
        additional_fields["asset"] = dataset["asset"]
    return additional_fields


def get_tags() -> list[SensorTag]:
    """
    The input tags for this model

    Returns
    -------
    list[SensorTag]
    """
    dataset = g.metadata["dataset"]
    tag_list = dataset["tag_list"]
    build_dataset_metadata = load_build_dataset_metadata()
    additional_fields = get_normalize_additional_fields(dataset)
    return normalize_sensor_tags(build_dataset_metadata, tag_list, **additional_fields)


def get_target_tags() -> list[SensorTag]:
    """
    The target tags for this model

    Returns
    -------
    list[SensorTag]
    """
    # TODO refactor this part to have the same tag preparation logic as in TimeSeriesDataset
    orig_target_tag_list = []
    if "target_tag_list" in g.metadata["dataset"]:
        orig_target_tag_list = g.metadata["dataset"]["target_tag_list"]
    if orig_target_tag_list:
        build_dataset_metadata = load_build_dataset_metadata()
        additional_fields = get_normalize_additional_fields(g.metadata["dataset"])
        return normalize_sensor_tags(
            build_dataset_metadata, orig_target_tag_list, **additional_fields
        )
    else:
        return get_tags()



# -*- coding: utf-8 -*-

from typing import List, Optional, Type, Any, cast
from copy import copy

from gordo.machine.validators import fix_runtime
from gordo.workflow.workflow_generator.helpers import patch_dict
from gordo.machine import (
    Machine,
    GlobalsConfig,
    load_machine_config,
    load_globals_config,
)
from gordo.utils import join_json_paths
from gordo import __version__
from gordo_core.import_utils import BackCompatibleLocations
from packaging.version import parse
from pydantic import parse_obj_as, BaseModel

from .schemas import BuilderPodRuntime, PodRuntime, Volume


def _calculate_influx_resources(nr_of_machines):
    return {
        "requests": {
            # The requests must be limited to keep the machine schedulable
            "memory": min(3000 + (220 * nr_of_machines), 28000),  # Between 3G and 28G
            "cpu": min(500 + (10 * nr_of_machines), 4000),  # Between 500m and 4000m
        },
        "limits": {
            "memory": min(3000 + (220 * nr_of_machines), 48000),
            "cpu": 10000 + (20 * nr_of_machines),
        },
    }


class NormalizedConfig:
    """
    Handles the conversion of a single Machine representation in config format
    and updates it with any features which are 'left out' inside of ``globals``
    key or the default config globals held here.
    """

    SPLITED_DOCKER_IMAGES: GlobalsConfig = {
        "runtime": {
            "deployer": {"image": "gordo-deploy"},
            "server": {"image": "gordo-model-server"},
            "prometheus_metrics_server": {"image": "gordo-model-server"},
            "builder": {"image": "gordo-model-builder"},
            "client": {"image": "gordo-client"},
        }
    }

    UNIFYING_GORDO_VERSION: str = "1.2.0"

    UNIFIED_DOCKER_IMAGES: GlobalsConfig = {
        "runtime": {
            "deployer": {"image": "gordo-base"},
            "server": {"image": "gordo-base"},
            "prometheus_metrics_server": {"image": "gordo-base"},
            "builder": {"image": "gordo-base"},
            "client": {"image": "gordo-base"},
        }
    }

    DEFAULT_CONFIG_GLOBALS: GlobalsConfig = {
        "runtime": {
            "reporters": [],
            "server": {
                "resources": {
                    "requests": {"memory": 3000, "cpu": 1000},
                    "limits": {"memory": 6000, "cpu": 2000},
                }
            },
            "prometheus_metrics_server": {
                "resources": {
                    "requests": {"memory": 200, "cpu": 100},
                    "limits": {"memory": 1000, "cpu": 200},
                }
            },
            "builder": {
                "resources": {
                    "requests": {"memory": 3900, "cpu": 1001},
                    "limits": {"memory": 31200, "cpu": 1001},
                },
                "remote_logging": {"enable": False},
            },
            "client": {
                "resources": {
                    "requests": {"memory": 3500, "cpu": 100},
                    "limits": {"memory": 4000, "cpu": 2000},
                },
                "max_instances": 30,
            },
            "influx": {"enable": True},
        },
        "evaluation": {
            "cv_mode": "full_build",
            "scoring_scaler": "sklearn.preprocessing.MinMaxScaler",
            "metrics": [
                "explained_variance_score",
                "r2_score",
                "mean_squared_error",
                "mean_absolute_error",
            ],
        },
    }

    def __init__(
        self,
        config: dict,
        project_name: str,
        gordo_version: Optional[str] = None,
        model_builder_env: Optional[dict] = None,
        back_compatibles: Optional[BackCompatibleLocations] = None,
        default_data_provider: Optional[str] = None,
        json_path: Optional[str] = None,
    ):
        if gordo_version is None:
            gordo_version = __version__
        default_globals = self.get_default_globals(gordo_version)
        default_globals["runtime"]["influx"][  # type: ignore
            "resources"
        ] = _calculate_influx_resources(  # type: ignore
            len(config["machines"])
        )

        passed_globals = load_globals_config(
            config.get("globals", dict()), join_json_paths("globals", json_path)
        )

        # keeping it for back-compatibility
        if model_builder_env is not None and not (
            default_globals
            and "runtime" in default_globals
            and "builder" in default_globals["runtime"]
            and "env" in default_globals["runtime"]["builder"]
        ):
            if "runtime" not in default_globals:
                default_globals["runtime"] = {}
            if "builder" not in default_globals["runtime"]:
                default_globals["runtime"]["builder"] = {}
            default_globals["runtime"]["builder"]["env"] = model_builder_env

        patched_globals = cast(
            GlobalsConfig,
            patch_dict(cast(dict, default_globals), cast(dict, passed_globals)),
        )
        patched_globals = self.prepare_patched_globals(patched_globals)

        self.project_name = project_name

        machines: list[Machine] = []
        for i, conf in enumerate(config["machines"]):
            machine_config = load_machine_config(
                conf, join_json_paths("machines[%d]" % i, json_path)
            )
            machine = Machine.from_config(
                cast(dict[str, Any], machine_config),
                project_name=project_name,
                config_globals=patched_globals,
                back_compatibles=back_compatibles,
                default_data_provider=default_data_provider,
            )
            machines.append(machine)

        self.machines = machines

        self.globals: GlobalsConfig = patched_globals

    @staticmethod
    def prepare_runtime(runtime: dict) -> dict:
        def prepare_pod_runtime(name: str, schema: Type[BaseModel] = PodRuntime):
            if name in runtime:
                # TODO handling pydantic.ValidationError
                pod_runtime = parse_obj_as(schema, runtime[name])
                runtime[name] = pod_runtime.dict(exclude_none=True)

        prepare_pod_runtime("builder", BuilderPodRuntime)
        if "volumes" in runtime:
            volumes = parse_obj_as(List[Volume], runtime["volumes"])
            runtime["volumes"] = [volume.dict(exclude_none=True) for volume in volumes]
        return runtime

    @classmethod
    def prepare_patched_globals(cls, patched_globals: GlobalsConfig) -> GlobalsConfig:
        runtime = fix_runtime(patched_globals.get("runtime"))
        runtime = cls.prepare_runtime(runtime)
        patched_globals["runtime"] = runtime
        return patched_globals

    @classmethod
    def get_default_globals(cls, gordo_version: str) -> GlobalsConfig:
        current_version = parse(gordo_version)
        unifying_version = parse(cls.UNIFYING_GORDO_VERSION)
        if current_version >= unifying_version:
            docker_images = cls.UNIFIED_DOCKER_IMAGES
        else:
            docker_images = cls.SPLITED_DOCKER_IMAGES
        default_globals = cls.DEFAULT_CONFIG_GLOBALS
        return cast(
            GlobalsConfig,
            patch_dict(cast(dict, copy(default_globals)), cast(dict, docker_images)),
        )



from subprocess import CompletedProcess

import pytest
from mock import patch
from packaging import version

from gordo.workflow.workflow_generator.helpers import (
    determine_argo_version,
    parse_argo_version,
    ArgoVersionError,
)

_argo_command = ["argo", "version", "--short"]


def test_parse_argo_version():
    parsed_version = parse_argo_version("2.12.11")
    assert type(parsed_version) is version.Version
    assert str(parsed_version) == "2.12.11"
    assert parse_argo_version("wrong_version") is None


def create_completed_process(return_code, stdout):
    return CompletedProcess(_argo_command, return_code, stdout=stdout, stderr=None)


def test_determine_argo_version_success():
    completed_process = create_completed_process(0, b"argo: v1.1.1\n")
    with patch("subprocess.run", return_value=completed_process):
        argo_version = determine_argo_version()
        assert str(argo_version) == "1.1.1"


def test_determine_argo_version_fail():
    with patch("subprocess.run", side_effect=FileNotFoundError("argo")):
        with pytest.raises(ArgoVersionError):
            determine_argo_version()
    with patch("subprocess.run", return_value=create_completed_process(0, b"\xa0\xa1")):
        with pytest.raises(ArgoVersionError):
            determine_argo_version()
    with patch(
        "subprocess.run", return_value=create_completed_process(0, b"wrong output")
    ):
        with pytest.raises(ArgoVersionError):
            determine_argo_version()



# -*- coding: utf-8 -*-
import glob
import importlib
import logging
import nbformat
import os
import sys

from nbconvert.exporters import PythonExporter

logger = logging.getLogger(__name__)


def test_notebooks(tmpdir):
    """
    Ensures all notebooks will run without error
    """
    repo_root = os.path.join(os.path.dirname(__file__), "..")
    notebooks = glob.glob(os.path.join(repo_root, "examples", "*.ipynb"))

    logger.info(f"Found {len(notebooks)} notebooks to test.")

    for notebook in notebooks:

        logger.info(f"Testing notebook: {os.path.basename(notebook)}")

        with open(notebook) as f:
            nb = nbformat.read(f, as_version=4)
            exporter = PythonExporter()
            source, _meta = exporter.from_notebook_node(nb)

            with open(os.path.join(tmpdir, "tmpmodule.py"), "w") as f:
                f.writelines(source)
            with open(os.path.join(tmpdir, "__init__.py"), "w") as f:
                f.write("from .tmpmodule import *")

            # Import this module to 'run' the code
            module_dir = os.path.join(tmpdir, "..")
            sys.path.insert(0, module_dir)

            importlib.import_module(os.path.basename(tmpdir), ".")

            sys.path.remove(module_dir)



# -*- coding: utf-8 -*-
import ipaddress
import re

import click
import json

from pydantic import parse_obj_as, ValidationError
from typing import Tuple, TypeVar, Generic, Type, Optional, Any

T = TypeVar("T")


class JSONParam(click.ParamType, Generic[T]):
    """
    Loads JSON and validates value against pydantic schema
    """

    name = "JSON"

    def __init__(self, schema: Type[T]):
        self.schema = schema

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Optional[T]:
        if value is None:
            return None
        try:
            data = json.loads(value)
        except json.JSONDecodeError as e:
            self.fail("Malformed JSON string - %s" % str(e))
        try:
            obj = parse_obj_as(self.schema, data)
        except ValidationError as e:
            self.fail("Schema validation error - %s" % str(e))
        return obj


class REParam(click.ParamType):
    """
    Validates argument over a regular expression
    """

    name = "REGEXP"

    def __init__(self, pattern: str):
        self.pattern = pattern
        self.re = re.compile(pattern)

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ):
        m = self.re.match(value)
        if not m:
            self.fail("Value '%s' not match '%s'" % (value, self.pattern))
        return value


class HostIP(click.ParamType):
    """
    Validate input is a valid IP address
    """

    name = "host"

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ):
        try:
            ipaddress.ip_address(value)
            return value
        except ValueError as e:
            self.fail(str(e))


def key_value_par(val) -> Tuple[str, str]:
    """
    Helper for CLI input of 'key,val'
    """
    return val.split(",")



import yaml

from copy import copy
from typing import TypedDict, Union, cast

from gordo_core.exceptions import ConfigException

from gordo.machine.metadata import Metadata
from gordo.utils import join_json_paths
from gordo_core.base import GordoBaseDataset

from .constants import MACHINE_YAML_FIELDS


class MachineConfigException(ConfigException):
    pass


class _BaseConfig(TypedDict, total=False):
    model: dict
    dataset: Union[dict, GordoBaseDataset]
    evaluation: dict
    metadata: Union[dict, Metadata]
    runtime: dict


class GlobalsConfig(_BaseConfig):
    pass


class MachineConfig(_BaseConfig):
    name: str


class ModelConfig(MachineConfig):
    project_name: str


def _load_config(config: dict, json_path: str = None) -> dict:
    new_config = copy(config)
    for field in config.keys():
        if field in MACHINE_YAML_FIELDS:
            val = config.get(field)
            if type(val) is str:
                try:
                    new_config[field] = yaml.safe_load(val)
                except yaml.YAMLError as e:
                    message = "Error loading YAML from '%s'" % join_json_paths(
                        field, json_path
                    )
                    raise MachineConfigException(message + ": " + str(e))
    return new_config


def load_globals_config(config: dict, json_path: str = None) -> GlobalsConfig:
    """
    Load `GlobalsConfig` from the dict

    Parameters
    ----------
    config: str
        Config to load.
    json_path: str
        JSON path position of the config.

    Returns
    -------

    """
    return cast(GlobalsConfig, _load_config(config, json_path))


def load_machine_config(config: dict, json_path: str = None) -> MachineConfig:
    """
    Load `MachineConfig` from the dict

    Parameters
    ----------
    config: str
        Config to load.
    json_path: str
        JSON path position of the config.

    Returns
    -------

    """
    machine_config = _load_config(config, json_path)
    if not machine_config.get("name"):
        raise MachineConfigException(
            "'%s' is empty" % join_json_paths("name", json_path)
        )
    return cast(MachineConfig, machine_config)


def load_model_config(config: dict, json_path: str = None) -> ModelConfig:
    """
    Load `ModelConfig` from the dict

    Parameters
    ----------
    config: str
        Config to load.
    json_path: str
        JSON path position of the config.

    Returns
    -------

    """
    model_config = cast(ModelConfig, load_machine_config(config, json_path))
    if not model_config.get("project_name"):
        raise MachineConfigException(
            "'%s' is empty" % join_json_paths("project_name", json_path)
        )
    return model_config



# -*- coding: utf-8 -*-
import json
import logging
from typing import Any, Optional, List, Callable, cast
from copy import copy

import yaml

from gordo_core.base import GordoBaseDataset
from gordo_core.sensor_tag import SensorTag
from gordo_core.import_utils import BackCompatibleLocations
from gordo.machine.validators import (
    ValidUrlString,
    ValidMetadata,
    ValidModel,
    ValidDataset,
    ValidMachineRuntime,
)
from gordo.machine.metadata import Metadata
from gordo.workflow.workflow_generator.helpers import patch_dict
from gordo.utils import normalize_sensor_tags, TagsList

from .constants import MACHINE_YAML_FIELDS
from .loader import ModelConfig, GlobalsConfig
from .encoders import MachineJSONEncoder, MachineSafeDumper, multiline_str

logger = logging.getLogger(__name__)


class Machine:
    """
    Represents a single machine in a config file
    """

    name = ValidUrlString()
    project_name = ValidUrlString()
    host = ValidUrlString()
    model = ValidModel()
    dataset = ValidDataset()
    metadata = ValidMetadata()
    runtime = ValidMachineRuntime()
    _strict = True

    @staticmethod
    def prepare_evaluation(evaluation: Optional[dict]) -> dict:
        if evaluation is None:
            evaluation = dict(cv_mode="full_build")
        return evaluation

    def __init__(
        self,
        name: str,
        model: dict,
        dataset: GordoBaseDataset,
        project_name: str,
        evaluation: Optional[dict] = None,
        metadata: Optional[Metadata] = None,
        runtime: Optional[dict] = None,
    ):

        if runtime is None:
            runtime = dict()
        if metadata is None:
            metadata = cast(Any, Metadata).from_dict({})
        self.name = name
        self.model = model
        self.dataset = dataset
        self.runtime = runtime
        self.evaluation = self.prepare_evaluation(evaluation)
        self.metadata = metadata
        self.project_name = project_name

        # host validation
        self.host = f"gordoserver-{self.project_name}-{self.name}"

    # TODO TypedDict for config argument
    @classmethod
    def from_config(  # type: ignore
        cls,
        config: dict[str, Any],
        project_name: Optional[str] = None,
        config_globals: GlobalsConfig = None,
        back_compatibles: Optional[BackCompatibleLocations] = None,
        default_data_provider: Optional[str] = None,
    ):
        """
        Construct an instance from a block of YAML config file which represents
        a single Machine; loaded as a ``dict``.

        Parameters
        ----------
        config: dict[str, Any]
            The loaded block of config which represents a 'Machine' in YAML
        project_name: str
            Name of the project this Machine belongs to.
        config_globals:
            The block of config within the YAML file within `globals`
        back_compatibles: Optional[BackCompatibleLocations]
            See `gordo_core.import_utils.prepare_back_compatible_locations()` function for reference.
        default_data_provider: Optional[str]

        Returns
        -------
        :class:`~Machine`
        """
        if config_globals is None:
            config_globals = dict()

        name = config["name"]
        config_model = config.get("model") or config_globals.get("model")
        if config_model is None:
            raise ValueError("model is empty")
        model = cast(dict, config_model)

        if project_name is None:
            project_name = config.get("project_name", None)
        if project_name is None:
            raise ValueError("project_name is empty")

        local_runtime = cast(dict, config.get("runtime", dict()))
        runtime = patch_dict(
            cast(dict, config_globals.get("runtime", dict())), local_runtime
        )

        dataset = patch_dict(
            config.get("dataset", dict()), config_globals.get("dataset", dict())
        )
        config_evaluation = cls.prepare_evaluation(config.get("evaluation"))
        evaluation = patch_dict(
            cast(dict, config_globals.get("evaluation", dict())), config_evaluation
        )

        metadata = Metadata(
            user_defined={
                "global-metadata": config_globals.get("metadata", dict()),
                "machine-metadata": config.get("metadata", dict()),
            }
        )
        return cls.from_dict(
            {
                "name": name,
                "model": model,
                "dataset": dataset,
                "project_name": project_name,
                "evaluation": evaluation,
                "metadata": metadata,
                "runtime": runtime,
            },
            back_compatibles=back_compatibles,
            default_data_provider=default_data_provider,
        )

    def normalize_sensor_tags(self, tag_list: TagsList) -> List[SensorTag]:
        """
        Finding assets for all of the tags according to information from the dataset metadata

        Parameters
        ----------
        tag_list: TagsList

        Returns
        -------
        List[SensorTag]

        """
        metadata = self.metadata
        build_dataset_metadata = metadata.build_metadata.dataset.to_dict()
        asset: Optional[str] = None
        if hasattr(self.dataset, "asset"):
            asset = self.dataset.asset
        return normalize_sensor_tags(build_dataset_metadata, tag_list, asset=asset)

    def __str__(self):
        return self.to_yaml()

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    # TODO TypedDict for d argument
    @classmethod
    def from_dict(
        cls,
        d: ModelConfig,
        back_compatibles: Optional[BackCompatibleLocations] = None,
        default_data_provider: Optional[str] = None,
    ) -> "Machine":
        """
        Create
        A dict taken from either gordo config or :func:`~Machine.to_dict`.
        """
        # No special treatment required, just here for consistency.
        d = copy(d)
        if "dataset" in d and isinstance(d["dataset"], dict):
            d["dataset"] = GordoBaseDataset.from_dict(
                d["dataset"],
                back_compatibles=back_compatibles,
                default_data_provider=default_data_provider,
            )
        if "metadata" in d and isinstance(d["metadata"], dict):
            d["metadata"] = cast(Any, Metadata).from_dict(d["metadata"])
        args = cast(dict, d)
        return cls(**args)

    def to_dict(self):
        """
        Convert to a ``dict`` representation along with all attributes which
        can also be converted to a ``dict``. Can reload with :func:`~Machine.from_dict`
        """
        return {
            "name": self.name,
            "dataset": self.dataset.to_dict(),
            "model": self.model,
            "metadata": self.metadata.to_dict(),
            "runtime": self.runtime,
            "project_name": self.project_name,
            "evaluation": self.evaluation,
        }

    def _to_yaml_dict(self, yaml_serializer: Callable[[Any], str] = None):
        if yaml_serializer is None:
            raise ValueError("yaml_serializer is empty")
        machine = self.to_dict()
        config = {}
        for k, v in machine.items():
            if k in MACHINE_YAML_FIELDS:
                v = yaml_serializer(v)
            config[k] = v
        return config

    def to_json(self):
        """
        Returns
        -------
            string JSON representation of the machine.
        """
        json_dumps: Callable[[Any], Any] = lambda v: json.dumps(
            v, cls=MachineJSONEncoder
        )
        return json_dumps(self._to_yaml_dict(json_dumps))

    def to_yaml(self):
        """
        Returns
        -------
            string YAML representation of the machine.
        """
        yaml_dump: Callable[[Any], Any] = lambda v: multiline_str(
            yaml.dump(v, Dumper=MachineSafeDumper)
        )
        return yaml.dump(self._to_yaml_dict(yaml_dump), Dumper=MachineSafeDumper)

    def report(self):
        """
        Run any reporters in the machine's runtime for the current state.

        Reporters implement the :class:`gordo.reporters.base.BaseReporter` and
        can be specified in a config file of the machine for example:

        .. code-block:: yaml

            runtime:
              reporters:
                - gordo.reporters.postgres.PostgresReporter:
                    host: my-special-host

        """
        # Avoid circular dependency with reporters which import Machine
        from gordo.reporters.base import BaseReporter

        for reporter in map(BaseReporter.from_dict, self.runtime.get("reporters", [])):
            logger.debug(f"Using reporter: {reporter}")
            reporter.report(self)



# -*- coding: utf-8 -*-

import simplejson
import logging
import os
import re
import pickle

from typing import Union, Any, Optional  # pragma: no flakes

from sklearn.pipeline import Pipeline
from sklearn.base import TransformerMixin, BaseEstimator  # noqa

from gordo.machine.model.base import GordoBase

logger = logging.getLogger(__name__)

N_STEP_REGEX = re.compile(r".*n_step=([0-9]+)")
CLASS_REGEX = re.compile(r".*class=(.*$)")


def dumps(model: Union[Pipeline, GordoBase]) -> bytes:
    """
    Dump a model into a bytes representation suitable for loading from
    ``gordo.serializer.loads``

    Parameters
    ----------
    model: Union[Pipeline, GordoBase]
        A gordo model/pipeline

    Returns
    -------
    bytes
        Serialized model which supports loading via ``serializer.loads()``

    Example
    -------
    >>> from gordo.machine.model.models import KerasAutoEncoder
    >>> from gordo import serializer
    >>>
    >>> model = KerasAutoEncoder('feedforward_symmetric')
    >>> serialized = serializer.dumps(model)
    >>> assert isinstance(serialized, bytes)
    >>>
    >>> model_clone = serializer.loads(serialized)
    >>> assert isinstance(model_clone, KerasAutoEncoder)
    """
    return pickle.dumps(model)


def loads(bytes_object: bytes) -> GordoBase:
    """
    Load a GordoBase model from bytes dumped from ``gordo.serializer.dumps``

    Parameters
    ----------
    bytes_object: bytes
        Bytes to be loaded, should be the result of `serializer.dumps(model)`

    Returns
    -------
    Union[GordoBase, Pipeline, BaseEstimator]
        Custom gordo model, scikit learn pipeline or other scikit learn like object.
    """
    return pickle.loads(bytes_object)


def metadata_path(
    source_dir: Union[os.PathLike, str]
) -> Optional[Union[os.PathLike, str]]:
    """
    Returns path to metadata.json file, if exists.

    """
    return _json_file_path(source_dir, "metadata.json")


def _json_file_path(source_dir: Union[os.PathLike, str], file_name: str):
    # Since this function can take the top level dir, or a dir directly
    # into the first step of the pipeline, we need to check both for the file
    possible_paths = [
        os.path.join(source_dir, file_name),
        os.path.join(source_dir, "..", file_name),
    ]
    return next((path for path in possible_paths if os.path.exists(path)), None)


def _load_json_file(source_dir: Union[os.PathLike, str], file_name: str) -> dict:
    file_path = _json_file_path(source_dir, file_name)

    if file_path:
        with open(file_path, "r") as f:
            return simplejson.load(f)
    else:
        raise FileNotFoundError(f"'{file_name}' file not found in '{source_dir}'")


def load_metadata(source_dir: Union[os.PathLike, str]) -> dict:
    """
    Load the given metadata.json which was saved during the ``serializer.dump``
    will return the loaded metadata as a dict, or empty dict if no file was found

    Parameters
    ----------
    source_dir: Union[os.PathLike, str]
        Directory of the saved model, As with serializer.load(source_dir) this
        source_dir can be the top level, or the first dir into the serialized model.

    Returns
    -------
    dict

    Raises
    ------
    FileNotFoundError
        If a 'metadata.json' file isn't found in or above the supplied ``source_dir``
    """
    return _load_json_file(source_dir, "metadata.json")


def load_info(source_dir: Union[os.PathLike, str]) -> dict:
    # TODO better docstring
    return _load_json_file(source_dir, "info.json")


def load(source_dir: Union[os.PathLike, str]) -> Any:
    """
    Load an object from a directory, saved by
    ``gordo.serializer.pipeline_serializer.dump``

    This take a directory, which is either top-level, meaning it contains
    a sub directory in the naming scheme: "n_step=<int>-class=<path.to.Class>"
    or the aforementioned naming scheme directory directly. Will return that
    unsterilized object.


    Parameters
    ----------
    source_dir: Union[os.PathLike, str]
        Location of the top level dir the pipeline was saved

    Returns
    -------
    Union[GordoBase, Pipeline, BaseEstimator]
    """
    # This source dir should have a single pipeline entry directory.
    # may have been passed a top level dir, containing such an entry:
    with open(os.path.join(source_dir, "model.pkl"), "rb") as f:
        return pickle.load(f)


def dump(
    obj: object,
    dest_dir: Union[os.PathLike, str],
    metadata: dict = None,
    info: Optional[dict] = None,
):
    """
    Serialize an object into a directory, the object must be pickle-able.

    Parameters
    ----------
    obj
        The object to dump. Must be pickle-able.
    dest_dir: Union[os.PathLike, str]
        The directory to which to save the model metadata: dict - any additional
        metadata to be saved alongside this model if it exists, will be returned
        from the corresponding "load" function
    metadata: Optional dict of metadata which will be serialized to a file together
        with the model, and loaded again by :func:`load_metadata`.
    info: Optional[str]
        Current revision info. For now, only used for storing "checksum"

    Returns
    -------
    None

    Example
    -------

    >>> from sklearn.pipeline import Pipeline
    >>> from sklearn.decomposition import PCA
    >>> from gordo.machine.model.models import KerasAutoEncoder
    >>> from gordo import serializer
    >>> from tempfile import TemporaryDirectory
    >>> pipe = Pipeline([
    ...     ('pca', PCA(3)),
    ...     ('model', KerasAutoEncoder(kind='feedforward_hourglass'))])
    >>> with TemporaryDirectory() as tmp:
    ...     serializer.dump(obj=pipe, dest_dir=tmp)
    ...     pipe_clone = serializer.load(source_dir=tmp)
    """
    with open(os.path.join(dest_dir, "model.pkl"), "wb") as m:
        pickle.dump(obj, m)
    if info is not None:
        with open(os.path.join(dest_dir, "info.json"), "w") as f:
            simplejson.dump(info, f, default=str)
    if metadata is not None:
        with open(os.path.join(dest_dir, "metadata.json"), "w") as f:
            simplejson.dump(metadata, f, default=str)



from typing import Optional, Any, List, Dict
from pydantic import BaseModel


class ResourceRequirements(BaseModel):
    requests: Optional[Dict[str, Any]]
    limits: Optional[Dict[str, Any]]


class ConfigMapKeySelector(BaseModel):
    key: str
    value: str


class SecretKeySelector(BaseModel):
    key: str
    name: str


class EnvVarSource(BaseModel):
    configMapKeyRef: Optional[ConfigMapKeySelector]
    secretKeyRef: Optional[SecretKeySelector]


class EnvVar(BaseModel):
    name: str
    value: Optional[str]
    valueFrom: Optional[EnvVarSource]


class ObjectMeta(BaseModel):
    labels: Optional[Dict[str, Any]]


class CSIVolumeSource(BaseModel):
    driver: str
    readOnly: bool = False
    fsType: Optional[str]
    volumeAttributes: Optional[Dict[str, Any]]


class Volume(BaseModel):
    name: str
    csi: Optional[CSIVolumeSource]


class VolumeMount(BaseModel):
    name: str
    readOnly: Optional[bool]
    mountPath: str


class PodRuntime(BaseModel):
    image: str
    resources: Optional[ResourceRequirements]
    metadata: Optional[ObjectMeta]
    env: Optional[List[EnvVar]]
    volumeMounts: Optional[List[VolumeMount]]


class RemoteLogging(BaseModel):
    enable: bool = False


class BuilderPodRuntime(PodRuntime):
    remote_logging: RemoteLogging


class Model(BaseModel):
    class Config:
        extra = "forbid"


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#capabilities-v1-core
class Capabilities(Model):
    add: Optional[List[str]]
    drop: Optional[List[str]]


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#selinuxoptions-v1-core
class SELinuxOptions(Model):
    level: Optional[str]
    role: Optional[str]
    type: Optional[str]
    user: Optional[str]


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#seccompprofile-v1-core
class SeccompProfile(Model):
    localhostProfile: Optional[str]
    type: Optional[str]


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#windowssecuritycontextoptions-v1-core
class WindowsSecurityContextOptions(Model):
    gmsaCredentialSpec: Optional[str]
    gmsaCredentialSpecName: Optional[str]
    runAsUserName: Optional[str]


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#securitycontext-v1-core
class SecurityContext(Model):
    allowPrivilegeEscalation: Optional[bool]
    capabilities: Optional[Capabilities]
    privileged: Optional[bool]
    procMount: Optional[str]
    readOnlyRootFilesystem: Optional[bool]
    runAsNonRoot: Optional[bool]
    runAsGroup: Optional[bool]
    runAsUser: Optional[int]
    seLinuxOptions: Optional[SELinuxOptions]
    seccompProfile: Optional[SeccompProfile]
    windowsOptions: Optional[WindowsSecurityContextOptions]


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#sysctl-v1-core
class Sysctl(Model):
    name: str
    value: str


# Reference https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#podsecuritypolicy-v1beta1-policy
class PodSecurityContext(Model):
    fsGroup: Optional[int]
    fsGroupChangePolicy: Optional[str]
    runAsGroup: Optional[int]
    runAsNonRoot: Optional[bool]
    runAsUser: Optional[int]
    seLinuxOptions: Optional[SELinuxOptions]
    seccompProfile: Optional[SeccompProfile]
    supplementalGroups: Optional[List[int]]
    sysctls: Optional[List[Sysctl]]
    windowsOptions: Optional[WindowsSecurityContextOptions]



import os
from pathlib import Path
from typing import Union, AnyStr, Optional

import logging

logger = logging.getLogger(__name__)

"""
A simple file-based key/value registry. Each key gets a file with filename = key, and
the content of the file is the value. No fancy. Why? Simple, and there is no problems
with concurrent writes to different keys. Concurrent writes to the same key will break 
stuff. 
"""


def write_key(registry_dir: Union[os.PathLike, str], key: str, val: AnyStr):
    """Registers a key-value combination into the register. Key must valid as a
    filename.

    Parameters
    ----------
    registry_dir: Union[os.PathLike, str]
        Path to the registry. If it does not exists it will be created, including any
        missing folders in the path.
    key: str
        Key to use for the key/value. Must be valid as a filename.
    val: AnyStr
        Value to write to the registry.

    Examples
    --------
    In the following example we use the temp directory as the registry
    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     write_key(tmpdir, "akey", "aval")
    ...     get_value(tmpdir, "akey")
    'aval'
    """
    key_file_path = Path(registry_dir).joinpath(key)
    logger.info(f"Storing object with key {key} to {key_file_path}")
    # If the model location already exists in the cache we overwrite it
    if key_file_path.exists():
        logger.warning(f"Key {key} already exists, " f"overwriting")
    elif not Path(registry_dir).exists():
        logger.debug(
            f"Found that registry dir {registry_dir} does not exists, creating it"
        )
        # If someone else creates the directory at the same time we accept that by
        # exists_ok=True
        os.makedirs(registry_dir, exist_ok=True)
    with key_file_path.open(mode="w") as f:
        f.write(val)  # type: ignore


def get_value(registry_dir: Union[os.PathLike, str], key: str) -> Optional[AnyStr]:
    """
    Retrieves the value with key reg_key from the registry, None if it does not
    exists.

    Parameters
    ----------
    registry_dir: Union[os.PathLike, str]
        Path to the registry. If it does not exist we return None
    key: str
        Key to look up in the registry.

    Returns
    -------
    Optional[AnyStr]:
        The value of `key` in the registry, None if no value is registered with that key
        in the registry.
    """
    output_val = None

    if registry_dir is None:
        return output_val

    key_file_path = Path(registry_dir).joinpath(key)
    logger.info(f"Looking for registry key {key} at path " f"{key_file_path}")
    # If the model location exists
    if key_file_path.exists():
        with key_file_path.open(mode="r") as f:
            output_val = f.read()
        logger.debug(f"Read value {output_val} from registry")
    else:
        logger.info(f"Did not find registry key {key} at path {key_file_path}")
    return output_val  # type: ignore


def delete_value(registry_dir: Union[os.PathLike, str], key: str) -> bool:
    """
    Deletes the value with key reg_key from the registry, and returns True if it
    existed.

    Parameters
    ----------
    registry_dir: Union[os.PathLike, str]
        Path to the registry. Does not need to exist
    key: str
        Key to look up in the registry.

    Returns
    -------
    bool:
        True if the key existed, false otherwise
    """
    key_file_path = Path(registry_dir).joinpath(key)
    logger.info(f"Looking for registry key {key} at path " f"{key_file_path}")
    # If the model location exists
    if key_file_path.exists():
        key_file_path.unlink()
        logger.debug(f"Removed key {key} from registry")
        return True
    else:
        logger.info(f"Did not find registry key {key} at path {key_file_path}")
    return False



#!/usr/bin/env python

import sys
import argparse
import re
import os
import json
import stat

from subprocess import Popen, PIPE
from packaging import version
from dataclasses import dataclass
from typing import List, Dict, Any, Optional, IO, cast

if sys.version_info.major != 3 or sys.version_info.minor < 7:
    raise RuntimeError("Unsupported python version: %s" % sys.version)

DOWNLOAD_URL = (
    "https://github.com/argoproj/argo-workflows/releases/download/v{version}/{arch}.gz"
)
ARCH = "argo-linux-amd64"
PROCESS_TIMEOUT = 60 * 5

BINARY_NAME = "argo"

_arch_re = re.compile(r"^[\w\-]+$")
_version_re = re.compile(r"^v?((\d+)\.(\d+)\.(\d+))")

_executable_mask = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH


@dataclass
class ArgoVersion:
    version: str
    number: Optional[int]
    is_default: bool

    _required_fields = ["version", "number"]

    @classmethod
    def load_list(cls, items: List[Dict[str, Any]]):
        is_default = True
        versions: List[ArgoVersion] = []
        numbers = set()
        for item in items:
            for field in cls._required_fields:
                if field not in item:
                    raise ValueError("'%s' is empty in %s" % (field, item))
            version, number = item["version"], int(item["number"])
            m = _version_re.match(version)
            if not m:
                raise ValueError("'%s' version is malformed" % version)
            if number in numbers:
                raise ValueError("Duplicates for version number %d", number)
            numbers.add(number)
            versions.append(cls(m[1], number, is_default))
            is_default = False
        return versions

    @property
    def binary_name(self):
        return BINARY_NAME + str(self.number)


def get_download_url(version: str, arch: str):
    return DOWNLOAD_URL.format(version=version, arch=arch)


def make_file_executable(file_path: str):
    s = os.stat(file_path)
    os.chmod(file_path, s.st_mode | _executable_mask)


def download_gz_binary(url: str, output_file: str, timeout: int = None):
    with open(output_file, "wb") as f:
        p1 = Popen(["curl", "-sL", url], stdout=PIPE)
        p2 = Popen(["gzip", "-d"], stdin=p1.stdout, stdout=f)
        cast(IO[bytes], p1.stdout).close()
        p2.communicate(timeout=timeout)
        if p2.returncode != 0:
            raise RuntimeError("Failed to download %s" % url)


def symlink(src: str, dst: str):
    if os.path.exists(dst):
        print("Removing %s" % dst)
        os.unlink(dst)
    print("Creating symlink %s -> %s" % (src, dst))
    os.symlink(src, dst)


def download_argo_versions(
    argo_versions: List[ArgoVersion], output_directory: str, arch: str, timeout: int
):
    for argo_version in argo_versions:
        url = get_download_url(argo_version.version, arch)
        output_file = os.path.join(output_directory, argo_version.binary_name)
        print("Downloading argo %s to %s" % (argo_version.version, output_file))
        download_gz_binary(url, output_file, timeout=timeout)
        print("Making %s executable" % output_file)
        make_file_executable(output_file)
        if argo_version.is_default:
            dst = os.path.join(output_directory, BINARY_NAME)
            symlink(argo_version.binary_name, dst)


def usage(
    parser: argparse.ArgumentParser, message: str = None, returncode: int = 1, f=None
):
    if f is None:
        f = sys.stdout
    if message:
        print(message + "\n", file=f)
    parser.print_help(file=f)
    sys.exit(returncode)


def main():
    parser = argparse.ArgumentParser(description="Download argo CLIs binaries")

    parser.add_argument(
        "-v",
        "--argo-versions",
        default=os.environ.get("ARGO_VERSIONS"),
        help='Argo versions list in JSON format. Takes ARGO_VERSIONS environment variable as default value. Example: [{"version": "3.4.2"}]',
    )
    parser.add_argument(
        "-o", "--output-directory", required=True, help="Output directory"
    )
    parser.add_argument(
        "--arch",
        default=ARCH,
        help="Binary architecture. Default: '%s'" % ARCH,
    )
    parser.add_argument(
        "-t",
        "--process-timeout",
        default=PROCESS_TIMEOUT,
        help="Subprocesses timeout in seconds. Default: %d" % PROCESS_TIMEOUT,
    )
    args = parser.parse_args()

    if not args.argo_versions:
        usage(parser, message="--argo-versions is empty")

    if not _arch_re.match(args.arch):
        raise ValueError("'%s' malformed arch" % args.arch)

    argo_versions = ArgoVersion.load_list(json.loads(args.argo_versions))

    download_argo_versions(
        argo_versions,
        output_directory=args.output_directory,
        arch=args.arch,
        timeout=args.process_timeout,
    )


if __name__ == "__main__":
    main()



from typing import Optional, Type

from gordo_core.import_utils import import_location

from .build_model import ModelBuilder


def create_model_builder(model_builder_class: Optional[str]) -> Type[ModelBuilder]:
    if model_builder_class is None:
        return ModelBuilder
    cls = import_location(model_builder_class)
    if not issubclass(cls, ModelBuilder):
        raise ValueError(
            '"%s" class located in "%s" should be subclass of "%s"'
            % (cls.__name__, model_builder_class, ModelBuilder.__name__)
        )
    return cls
